@@ -1,5 +1,5 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2017-2022 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2017-2023 Brett Sheffield <bacs@librecast.net> */
 /* librecast/net.h - librecast network API */
 
 #ifndef _LIBRECAST_NET_H
@@ -11,11 +11,17 @@ extern int (*lc_msg_logger)(lc_channel_t *, lc_message_t *, void *logdb);
 
 /* create new librecast context and set up environment
  * call lc_ctx_free() when done */
-lc_ctx_t *lc_ctx_new();
+lc_ctx_t *lc_ctx_new(void);
 
 /* destroy librecast context and clean up */
 void lc_ctx_free(lc_ctx_t *ctx);
 
+/* set debug level */
+int lc_ctx_debug(lc_ctx_t *ctx, int debug);
+
+/* set output stream for debug */
+FILE *lc_ctx_stream(lc_ctx_t *ctx, FILE *stream);
+
 /* create librecast socket */
 lc_socket_t *lc_socket_new(lc_ctx_t *ctx);
 
@@ -55,6 +61,25 @@ int lc_channel_join(lc_channel_t *chan);
 /* leave a librecast channel */
 int lc_channel_part(lc_channel_t *chan);
 
+/* set default interface for sockets and channels created with this context */
+void lc_ctx_ifx(lc_ctx_t *ctx, unsigned int ifx);
+
+/* set ratelimit on ctx - this will be applied to all subsequently created
+ * channels on that ctx */
+void lc_ctx_ratelimit(lc_ctx_t *ctx, size_t bps_out, size_t bps_in);
+
+/* set channel rate limits */
+void lc_channel_ratelimit(lc_channel_t *chan, size_t bps_out, size_t bps_in);
+
+/* set default coding for sockets and channels created with this context */
+int lc_ctx_coding_set(lc_ctx_t *ctx, int coding);
+
+/* set symmetric key for context */
+int lc_ctx_getkey(lc_ctx_t *ctx, lc_key_t *key, int type);
+int lc_ctx_setkey(lc_ctx_t *ctx, lc_key_t *key, int type);
+int lc_ctx_set_sym_key(lc_ctx_t *ctx, unsigned char *key, size_t len);
+int lc_ctx_set_pub_key(lc_ctx_t *ctx, unsigned char *key, size_t len);
+
 /* set channel coding */
 int lc_channel_coding_set(lc_channel_t *chan, int coding);
 
@@ -157,13 +182,6 @@ int lc_msg_get(lc_message_t *msg, lc_msg_attr_t attr, void **value);
 /* set message attributes */
 int lc_msg_set(lc_message_t *msg, lc_msg_attr_t attr, void *value);
 
-/* access structure internals */
-
-/* return structure ids */
-uint32_t lc_ctx_get_id(lc_ctx_t *ctx);
-uint32_t lc_socket_get_id(lc_socket_t *sock);
-uint32_t lc_channel_get_id(lc_channel_t *chan);
-
 /* return raw network socket */
 int lc_socket_raw(lc_socket_t *sock);
 int lc_channel_socket_raw(lc_channel_t *chan);
@@ -171,6 +189,14 @@ int lc_channel_socket_raw(lc_channel_t *chan);
 /* return context for channel chan */
 lc_ctx_t *lc_channel_ctx(lc_channel_t *chan);
 
+#ifdef HAVE_LIBLCRQ
+#include <lcrq.h>
+/* return lcrq handle */
+rq_t *lc_channel_rq(lc_channel_t *chan);
+
+int lc_channel_oti_peek(lc_channel_t *chan, rq_oti_t *oti, rq_scheme_t *scheme);
+#endif
+
 /* return socket bound to this channel */
 lc_socket_t *lc_channel_socket(lc_channel_t *chan);
 
@@ -183,6 +209,13 @@ struct in6_addr *lc_channel_in6addr(lc_channel_t *chan);
 /* return channel uri */
 char *lc_channel_uri(lc_channel_t *chan);
 
+/* create IPv6 multicast addr from supplied hash and flags */
+void lc_hashtoaddr(struct in6_addr *addr, unsigned char *hash, size_t hashlen, unsigned char flags);
+
+/* Create a new channel using supplied hash, flags and port */
+lc_channel_t * lc_channel_hash(lc_ctx_t *ctx, unsigned char *hash, size_t hashlen,
+		unsigned char flags, short port);
+
 /* create new channel from grp address and service */
 lc_channel_t * lc_channel_init(lc_ctx_t *ctx, struct sockaddr_in6 *sa);
 
