@@ -29,12 +29,18 @@
 # define OS_NETBSD 1
 # include <netinet6/ip6_mroute.h>
 #endif
-#if OS_NETBSD
-#include <fcntl.h>
-#include <net/bpf.h>
-#include <sys/ioctl.h>
-#define sock_filter bpf_insn
-#define sock_fprog  bpf_program
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
+#endif
+#ifdef HAVE_NET_BPF_H
+# include <net/bpf.h>
+# ifndef sock_filter
+#  define sock_filter bpf_insn
+#  define sock_fprog  bpf_program
+# endif
+#endif
+#ifdef HAVE_SYS_IOCTL_H
+# include <sys/ioctl.h>
 #endif
 
 #include <arpa/inet.h>
@@ -51,18 +57,40 @@
 #include <string.h>
 #include <unistd.h>
 
-#ifdef HAVE_SYS_ENDIAN_H
-#include <sys/endian.h>
+#ifdef HAVE_ENDIAN_H
+# include <endian.h>
+#elif defined(HAVE_SYS_ENDIAN_H)
+# include <sys/endian.h>
+#elif defined(HAVE_LIBKERN_OSBYTEORDER_H)
+# include <libkern/OSByteOrder.h>
+# define be64toh(x) OSSwapBigToHostInt64(x)
 #endif
 
-// FIXME -> configure
-#define HAVE_NETLINK __linux__
-
 /* BPF filter for MLD types - we are interested in icmpv6 types 130-132 and 143 */
-/* `tcpdump -dd "ip6[48] == 130 or ip6[48] == 131 or ip6[48] == 132 or ip6[48] == 143"` */
+/* tcpdump -dd 'ip6[6] == 0 and
+ *    ((ip6[40:4] == 0x3a000502 and ip6[44:2] == 0) or
+ *     (ip6[40:4] == 0x3a000100 and ip6[44:4] == 0x05020000)) and
+ * (ip6[48] == 130 or ip6[48] == 131 or ip6[48] == 132 or ip6[48] == 143)'
+ * notes:
+ * ip6[6] == 0: first header is an option header
+ *   ip6[40:4] == 0x3a000502: router alert, next header ICMP6 and Option header 0
+ *   ip6[44:2] == 0: alert type MLD
+ * -or-
+ *   ip6[40:4] == 0x3a000100: router alert, next header ICMP6 and Option header 1
+ *   ip6[44:4] == 0x05020000: alert type MLD
+ * ip6[48]: ICMP6 type in ICMP6 header */
 static struct sock_filter BPF_MLD[] = {
 	{ 0x28, 0, 0, 0x0000000c },
-	{ 0x15, 0, 6, 0x000086dd },
+	{ 0x15, 0, 15, 0x000086dd },
+	{ 0x30, 0, 0, 0x00000014 },
+	{ 0x15, 0, 13, 0x00000000 },
+	{ 0x20, 0, 0, 0x00000036 },
+	{ 0x15, 0, 2, 0x3a000502 },
+	{ 0x28, 0, 0, 0x0000003a },
+	{ 0x15, 3, 9, 0x00000000 },
+	{ 0x15, 0, 8, 0x3a000100 },
+	{ 0x20, 0, 0, 0x0000003a },
+	{ 0x15, 0, 6, 0x05020000 },
 	{ 0x30, 0, 0, 0x0000003e },
 	{ 0x15, 3, 0, 0x00000082 },
 	{ 0x15, 2, 0, 0x00000083 },
@@ -99,9 +127,7 @@ static void freeifacelist(mld_iface_t *i)
 
 static mld_iface_t *mld_get_iface(const mld_t *mld, const unsigned int ifx)
 {
-	int ifaces = 0;
 	for (mld_iface_t *i = mld->iface; i; i = i->next) {
-		ifaces++;
 		if (i->ifx == ifx) return i;
 	}
 	return NULL;
@@ -122,11 +148,16 @@ mld_grp_list_t *mld_get_grp(const mld_t *mld, const unsigned int ifx, const stru
 
 void mld_free(mld_t *mld)
 {
+	int err = errno;
 	freewatchlist(mld->watch);
 	freeifacelist(mld->iface);
 	sem_destroy(&mld->sem_mld);
 	sem_destroy(&mld->sem_state);
+#if HAVE_NETLINK
+	sem_destroy(&mld->sem_netlink);
+#endif
 	free(mld);
+	errno = err;
 }
 
 int mld_del_iface(mld_t *mld, unsigned int ifx)
@@ -169,7 +200,7 @@ static int mld_join_iface(mld_t *mld, unsigned int ifx)
 	}
 	req.ipv6mr_interface = ifx;
 	if (getifaddrs(&ifaddr)) {
-		ERROR("%s() getifaddrs: %s", __func__, strerror(errno));
+		ERROR("%s() getifaddrs: %s\n", __func__, strerror(errno));
 		return -1;
 	}
 	/* find link-local address for ifx and bind to perform JOIN */
@@ -188,11 +219,11 @@ static int mld_join_iface(mld_t *mld, unsigned int ifx)
 			continue;
 		}
 		if (!setsockopt(mld->sock, IPPROTO_IPV6, IPV6_JOIN_GROUP, &req, sizeof(req))) {
-			DEBUG("MLD listening on interface %s[%u]", ifa->ifa_name, ifx);
+			DEBUG("MLD listening on interface %s[%u]\n", ifa->ifa_name, ifx);
 			memcpy(&iface->llink, llocal, sizeof(struct sockaddr_in6));
 		}
 		else {
-			ERROR("%s() setsockopt: %s", __func__, strerror(errno));
+			ERROR("%s() setsockopt: %s\n", __func__, strerror(errno));
 		}
 	}
 	freeifaddrs(ifaddr);
@@ -228,18 +259,21 @@ exit_1:
 int mld_add_iface(mld_t *mld, unsigned int ifx, char *ifname)
 {
 	mld_iface_t *iface;
-	if (mld_get_iface(mld, ifx)) return -1;
+	if (strlen(ifname) >= 16)
+		return (errno = ENAMETOOLONG), -1;
+	if (mld_get_iface(mld, ifx))
+		return (errno = ENODEV), -1;
 	iface = calloc(1, sizeof(mld_iface_t));
+	if (!iface) return -1;
 	iface->next = mld->iface;
 	iface->ifx = ifx;
-	assert(strlen(ifname) < 16);
 	strcpy(iface->ifname, ifname);
 #if OS_NETBSD
 	iface->bpf = mld_bpf_init(ifname);
 #endif
 	mld->iface = iface;
 	mld_join_iface(mld, ifx);
-	DEBUG("ifx %u added", ifx);
+	DEBUG("ifx %u added\n", ifx);
 	mld_watch_event_push(mld, ifx, NULL, MLD_EVENT_IFUP);
 	return 0;
 }
@@ -248,7 +282,7 @@ int mld_add_iface_ifx(mld_t *mld, unsigned int ifx)
 {
 	char ifname[IF_NAMESIZE];
 	if (!if_indextoname(ifx, ifname)) {
-		ERROR("%s() if_indextoname: %s", __func__, strerror(errno));
+		ERROR("%s() if_indextoname: %s\n", __func__, strerror(errno));
 		return -1;
 	}
 	return mld_add_iface(mld, ifx, ifname);
@@ -292,7 +326,7 @@ void *mld_state_ifx_add_all(void *arg)
 	struct ifaddrs *ifa = NULL, *ifap = NULL;
 	unsigned int ifx;
 	if (getifaddrs(&ifa) == -1) {
-		ERROR("%s() getifaddrs(): %s", __func__, strerror(errno));
+		ERROR("%s() getifaddrs(): %s\n", __func__, strerror(errno));
 		return NULL;
 	}
 	for (ifap = ifa; ifap; ifap = ifap->ifa_next) {
@@ -340,36 +374,55 @@ static void *mld_state_grp_del(void *arg)
 	return arg;
 }
 
-static void *mld_state_grp_cmp(void *arg)
+static int mld_state_grp_check(struct mld_txn_s *txn)
 {
-	struct mld_txn_s *txn = arg;
 	mld_grp_list_t *grp;
-	mld_iface_t *iface = mld_get_iface(txn->mld, txn->ifx);
-	if (!iface) {
-		txn->err = ENODEV;
-		txn->ret = -1;
-		return arg;
-	}
-	grp = mld_get_grp(txn->mld, txn->ifx, txn->addr);
-	if (grp) {
+	if ((grp = mld_get_grp(txn->mld, txn->ifx, txn->addr))) {
 		struct timespec ts = {0};
-		clock_gettime(CLOCK_REALTIME, &ts);
+		mld_iface_t *iface = mld_get_iface(txn->mld, txn->ifx);
+		if (!iface) return (errno = ENODEV), -1;
+		if (clock_gettime(CLOCK_REALTIME, &ts) == -1) return -1;
 		if (grp->expires.tv_sec <= ts.tv_sec) { /* expired, delete */
-			txn->ret = mld_state_grp_del_entry(iface, txn->addr);
-			txn->err = errno;
+			return mld_state_grp_del_entry(iface, txn->addr);
 		}
 		else {
 			txn->expires.tv_sec = grp->expires.tv_sec;
 			txn->expires.tv_nsec = grp->expires.tv_nsec;
-			txn->ret = 1;
+			return 1;
 		}
 	}
+	return 0;
+}
+
+static void *mld_state_grp_cmp(void *arg)
+{
+	struct mld_txn_s *txn = arg;
+	mld_iface_t *iface;
+	if (!txn->ifx) {
+		/* check all interfaces, return 1 if *any* match */
+		txn->ret = 0;
+		for (mld_iface_t *i = txn->mld->iface; i; i = i->next) {
+			txn->ifx = i->ifx;
+			txn->ret = mld_state_grp_check(txn);
+			if (txn->ret) break;
+		}
+	}
+	else if (!(iface = mld_get_iface(txn->mld, txn->ifx))) {
+		txn->err = ENODEV;
+		txn->ret = -1;
+	}
+	else {
+		txn->ret = mld_state_grp_check(txn);
+		txn->err = errno;
+	}
 	return arg;
 }
 
 static void *mld_state_grp_add(void *arg)
 {
-	struct mld_txn_s *txn = arg;
+	struct mld_txn_s *txn;
+	if (!arg) return NULL;
+	txn = arg;
 	mld_grp_list_t *grp;
 	mld_iface_t *iface = mld_get_iface(txn->mld, txn->ifx);
 	if (!iface) {
@@ -395,7 +448,7 @@ static void *mld_state_grp_add(void *arg)
 	grp->expires.tv_sec = txn->expires.tv_sec;
 	grp->expires.tv_nsec = txn->expires.tv_nsec;
 
-#ifdef DEBUG_ON
+#if DEBUG_ON
 	char strgrp[INET6_ADDRSTRLEN];
 	char ifname[IF_NAMESIZE];
 	if_indextoname(txn->ifx, ifname);
@@ -472,7 +525,9 @@ void mld_wait_callback(mld_watch_t *watch)
 
 void *mld_wait_callback_thread(void *arg)
 {
-	mld_watch_t *watch = (mld_watch_t *)arg;
+	mld_watch_t *watch;
+	if (!arg) return NULL;
+	watch = (mld_watch_t *)arg;
 	watch->f((mld_watch_t *)arg);
 	free(arg);
 	return NULL;
@@ -488,18 +543,26 @@ int mld_wait(mld_t *mld, unsigned int ifx, struct in6_addr *addr, int flags)
 		errno = EWOULDBLOCK;
 		return -1;
 	}
-	sem_init(&sem_watch, 0, 0);
-	mld_watch_add(mld, ifx, addr, &mld_wait_callback, &sem_watch, 0);
-	sem_wait(&sem_watch);
+	if (sem_init(&sem_watch, 0, 0)) return -1;
+	if (mld_watch_add(mld, ifx, addr, &mld_wait_callback, &sem_watch, 0)) {
+		rc = sem_wait(&sem_watch);
+	}
+	else rc = -1;
 	sem_destroy(&sem_watch);
-	return 0;
+	return rc;
 }
 
 /* check for watchers, run callbacks in detached thread */
 static void *mld_watch_event(void *arg)
 {
-	mld_txn_t *txn = (mld_txn_t *)arg;
+	pthread_attr_t attr;
+	mld_txn_t *txn;
 	mld_watch_t *event;
+	if (!arg) return NULL;
+	if (pthread_attr_init(&attr)) return NULL;
+	if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED))
+		goto err_attr_destroy;
+	txn = (mld_txn_t *)arg;
 	for (mld_watch_t *w = txn->mld->watch; w; w = w->next) {
 		if (w->ifx && txn->ifx && txn->ifx != w->ifx) continue;
 		if (w->grp.s6_addr[0] && txn->addr && memcmp(txn->addr, &w->grp, sizeof(struct in6_addr)))
@@ -510,11 +573,13 @@ static void *mld_watch_event(void *arg)
 		event->flags = txn->type;
 		event->ifx = txn->ifx;
 		if (txn->addr) memcpy(&event->grp, txn->addr, sizeof(struct in6_addr));
-		pthread_create(&w->tid, NULL, &mld_wait_callback_thread, event);
-		pthread_detach(w->tid);
+		if (pthread_create(&w->tid, &attr, &mld_wait_callback_thread, event))
+			free(event);
 	}
 	free(txn->addr);
 	free(txn);
+err_attr_destroy:
+	pthread_attr_destroy(&attr);
 	return NULL;
 }
 
@@ -525,15 +590,20 @@ static inline void mld_watch_event_push(mld_t *mld, unsigned int ifx, struct in6
 	mld_txn_t *txn;
 	if (!mld->q[MLD_WATCH]) return;
 	txn = calloc(1, sizeof (mld_txn_t));
+	if (!txn) return;
 	if (addr) {
 		grp = malloc(sizeof(struct in6_addr));
+		if (!grp) goto err_free_txn;
 		memcpy(grp, addr, sizeof(struct in6_addr));
 	}
 	txn->mld = mld;
 	txn->ifx = ifx;
 	txn->addr = grp;
 	txn->type = type;
-	job_push_new(mld->q[MLD_WATCH], &mld_watch_event, txn, sizeof(*txn), &free, 0);
+	job_push_new(mld->q[MLD_WATCH], &mld_watch_event, txn, sizeof(*txn), NULL, 0);
+	return;
+err_free_txn:
+	free(txn);
 }
 
 /* check every grp on every interface for expiry */
@@ -541,7 +611,7 @@ static void *mld_state_timer_check(void *arg)
 {
 	mld_t *mld = (mld_t *)arg;
 	struct timespec now = {0};
-	clock_gettime(CLOCK_REALTIME, &now);
+	if (clock_gettime(CLOCK_REALTIME, &now) == -1) return NULL;
 	for (mld_iface_t *i = mld->iface; i; i = i->next) {
 		for (mld_grp_list_t *g = i->grp, *p = NULL; g;) {
 			if (g->expires.tv_sec <= now.tv_sec) {
@@ -569,24 +639,49 @@ static void *mld_state_timer_check(void *arg)
 	return arg;
 }
 
+static void txn_done(void *arg)
+{
+	struct mld_txn_s *txn = (struct mld_txn_s *)arg;
+	sem_post(&txn->done);
+}
+
+static int txn_init(struct mld_txn_s *txn, mld_t *mld, unsigned int ifx, struct in6_addr *addr,
+		struct timespec *ts)
+{
+	if (sem_init(&txn->done, 0, 0)) return -1;
+	txn->mld = mld;
+	txn->ifx = ifx;
+	txn->addr = addr;
+	if (ts) {
+		txn->expires.tv_sec = ts->tv_sec;
+		txn->expires.tv_nsec = ts->tv_nsec;
+	}
+	return 0;
+}
+
+static int txn_wait(struct mld_txn_s *txn)
+{
+	if (sem_wait(&txn->done) || sem_destroy(&txn->done)) return -1;
+	errno = txn->err;
+	return txn->ret;
+}
+
 int mld_filter_ifx_del(mld_t *mld, unsigned int ifx)
 {
 	struct mld_txn_s txn = {0};
-	txn.mld = mld;
-	txn.ifx = ifx;
-	job_push_new(mld->q[MLD_STATE], &mld_state_ifx_del, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
-	return txn.ret;
+	if (txn_init(&txn, mld, ifx, NULL, NULL) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_ifx_del, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	return txn_wait(&txn);
 }
 
 int mld_filter_ifx_add(mld_t *mld, unsigned int ifx)
 {
 	struct mld_txn_s txn = {0};
-	txn.mld = mld;
-	txn.ifx = ifx;
-	job_push_new(mld->q[MLD_STATE], &mld_state_ifx_add, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
-	return txn.ret;
+	if (txn_init(&txn, mld, ifx, NULL, NULL) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_ifx_add, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	return txn_wait(&txn);
 }
 
 int mld_filter_grp_cmp(mld_t *mld, unsigned int ifx, struct in6_addr *addr)
@@ -597,12 +692,10 @@ int mld_filter_grp_cmp(mld_t *mld, unsigned int ifx, struct in6_addr *addr)
 		return -1;
 	}
 	/* push job to STATE queue */
-	txn.mld = mld;
-	txn.ifx = ifx;
-	txn.addr = addr;
-	job_push_new(mld->q[MLD_STATE], &mld_state_grp_cmp, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
-	return txn.ret;
+	if (txn_init(&txn, mld, ifx, addr, NULL) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_grp_cmp, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	return txn_wait(&txn);
 }
 
 int mld_filter_timer_get(mld_t *mld, unsigned int ifx, struct in6_addr *addr, struct timespec *ts)
@@ -613,11 +706,10 @@ int mld_filter_timer_get(mld_t *mld, unsigned int ifx, struct in6_addr *addr, st
 		return -1;
 	}
 	/* push job to STATE queue */
-	txn.mld = mld;
-	txn.ifx = ifx;
-	txn.addr = addr;
-	job_push_new(mld->q[MLD_STATE], &mld_state_grp_cmp, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
+	if (txn_init(&txn, mld, ifx, addr, ts) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_grp_cmp, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	txn_wait(&txn);
 	ts->tv_sec = txn.expires.tv_sec;
 	ts->tv_nsec = txn.expires.tv_nsec;
 	return txn.ret;
@@ -634,12 +726,10 @@ int mld_filter_grp_del(mld_t *mld, unsigned int ifx, struct in6_addr *addr)
 	}
 
 	/* push job to STATE queue */
-	txn.mld = mld;
-	txn.ifx = ifx;
-	txn.addr = addr;
-	job_push_new(mld->q[MLD_STATE], &mld_state_grp_del, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
-	return txn.ret;
+	if (txn_init(&txn, mld, ifx, addr, NULL) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_grp_del, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	return txn_wait(&txn);
 }
 
 int mld_filter_timer_set(mld_t *mld, unsigned int ifx, struct in6_addr *addr, struct timespec *ts)
@@ -651,23 +741,17 @@ int mld_filter_timer_set(mld_t *mld, unsigned int ifx, struct in6_addr *addr, st
 		errno = EINVAL;
 		return -1;
 	}
-
-	txn.mld = mld;
-	txn.ifx = ifx;
-	txn.addr = addr;
-	txn.expires.tv_sec = ts->tv_sec;
-	txn.expires.tv_nsec = ts->tv_nsec;
-
 	/* push job to STATE queue to update timer */
-	job_push_new(mld->q[MLD_STATE], &mld_state_grp_add, &txn, sizeof(txn), &free, JOB_WAIT);
-	errno = txn.err;
-	return txn.ret;
+	if (txn_init(&txn, mld, ifx, addr, ts) == -1) return -1;
+	if (!job_push_new(mld->q[MLD_STATE], &mld_state_grp_add, &txn, sizeof(txn), &txn_done, 0))
+		return -1;
+	return txn_wait(&txn);
 }
 
 int mld_filter_timer_set_s(mld_t *mld, unsigned int ifx, struct in6_addr *addr, int s)
 {
 	struct timespec ts = {0};
-	clock_gettime(CLOCK_REALTIME, &ts);
+	if (clock_gettime(CLOCK_REALTIME, &ts) == -1) return -1;
 	ts.tv_sec += s;
 	return mld_filter_timer_set(mld, ifx, addr, &ts);
 }
@@ -687,7 +771,7 @@ static int mld_is_querier(mld_t *mld, unsigned int ifx, struct timespec *ts)
 	struct timespec now;
 	if (!iface) return 0;
 	if (ts) now.tv_sec = ts->tv_sec;
-	else clock_gettime(CLOCK_REALTIME, &now);
+	else if (clock_gettime(CLOCK_REALTIME, &now) == -1) return -1;
 	return (iface->qseen.tv_sec + MLD2_OTHER_QUERIER_TIMEOUT > now.tv_sec) ? 0 : 1;
 }
 
@@ -701,19 +785,19 @@ static int mld_query_cmsglen(mld_msg_t *msg)
 {
 	int rc;
 	if ((rc = inet6_opt_init(NULL, 0)) == -1) {
-		ERROR("inet6_opt_init failed");
+		ERROR("inet6_opt_init failed\n");
 		return -1;
 	}
 	msg->extlen = (socklen_t)rc;
 	assert(msg->extlen > 0);
 	if ((rc = inet6_opt_append(NULL, 0, msg->extlen, IP6OPT_ROUTER_ALERT, 2, 2, NULL)) == -1) {
-		ERROR("inet6_opt_append failed");
+		ERROR("inet6_opt_append failed\n");
 		return -1;
 	}
 	msg->extlen = (socklen_t)rc;
 	assert(msg->extlen > 0);
 	if ((rc = inet6_opt_finish(NULL, 0, msg->extlen)) == -1) {
-		ERROR("inet6_opt_finish failed");
+		ERROR("inet6_opt_finish failed\n");
 		return -1;
 	}
 	msg->extlen = (socklen_t)rc;
@@ -754,18 +838,18 @@ int mld_query_msg(mld_t *mld, unsigned int ifx, struct in6_addr *saddr, mld_msg_
 	extbuf = CMSG_DATA(msg->cmsgh);
 	if ((offset = inet6_opt_init(extbuf, msg->extlen)) == -1) {
 		ERROR("inet6_opt_init");
-		return -1;
+		goto err_free_cmsgbuf;
 	}
 	if ((offset = inet6_opt_append(extbuf, msg->extlen, offset,
 		IP6OPT_ROUTER_ALERT, 2, 2, &databufp)) == -1)
 	{
-		ERROR("inet6_opt_append");
-		return -1;
+		ERROR("inet6_opt_append\n");
+		goto err_free_cmsgbuf;
 	}
 	inet6_opt_set_val(databufp, 0, &racode, sizeof(racode));
 	if (inet6_opt_finish(extbuf, msg->extlen, offset) == -1) {
-		ERROR("inet6_opt_finish");
-		return -1;
+		ERROR("inet6_opt_finish\n");
+		goto err_free_cmsgbuf;
 	}
 	if (saddr) {
 		/* Multicast Address Specific Query */
@@ -776,12 +860,15 @@ int mld_query_msg(mld_t *mld, unsigned int ifx, struct in6_addr *saddr, mld_msg_
 		/* MLDv2 General Query */
 		memset(&msg->dst.sin6_addr, 0, sizeof(struct in6_addr));
 		if (inet_pton(AF_INET6, MLD2_ALL_NODES, &msg->dst.sin6_addr) != 1) {
-			ERROR("inet_pton");
-			return -1;
+			ERROR("inet_pton\n");
+			goto err_free_cmsgbuf;
 		}
 		memcpy(&msg->qmsg.addr, &msg->dst.sin6_addr, sizeof(struct in6_addr));
 	}
 	return 0;
+err_free_cmsgbuf:
+	free(msg->cmsgbuf);
+	return -1;
 }
 
 static void mld_query_send(mld_t *mld, unsigned int ifx, struct in6_addr *saddr)
@@ -796,7 +883,7 @@ static void mld_query_send(mld_t *mld, unsigned int ifx, struct in6_addr *saddr)
 	if (!iface) return;
 	if (saddr) {
 		/* Multicast Address Specific Query - check timer */
-		clock_gettime(CLOCK_REALTIME, &now);
+		if (clock_gettime(CLOCK_REALTIME, &now) == -1) return;
 		/* check if we are the Querier (for General queries we've already done this) */
 		if (!mld_is_querier(mld, ifx, &now)) return;
 		mld_grp_list_t *g;
@@ -808,28 +895,38 @@ static void mld_query_send(mld_t *mld, unsigned int ifx, struct in6_addr *saddr)
 	}
 	if (mld_query_msg(mld, ifx, saddr, &msg) == -1) return;
 	if ((setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))) == -1) {
-		ERROR("setsockopt(SO_REUSEADDR)");
+		ERROR("setsockopt(SO_REUSEADDR)\n");
 	}
+#ifdef SO_REUSEPORT
 	if ((setsockopt(sock, SOL_SOCKET, SO_REUSEPORT, &opt, sizeof(opt))) == -1) {
-		ERROR("setsockopt(SO_REUSEPORT)");
+		ERROR("setsockopt(SO_REUSEPORT)\n");
 	}
+#endif
+#ifdef SO_BINDTODEVICE
+	/* if a previous call had a bind() and ifx is different from what it had
+	 * at the time, the next setsockopt(IPV6_MULTICAST_IF) will fail; this
+	 * works around it */
+	opt = 0;
+	if (setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, &opt, sizeof(opt)) == -1)
+		ERROR("setsockopt(SO_BINDTODEVICE)\n");
+#endif
 	if (saddr) {
 		/* bind iff Address Specific Query and address is NOT local scope
 		 * FreeBSD fails if we bind() when sending to local scope */
 		if (!IN6_IS_ADDR_MC_LINKLOCAL(saddr)) {
 			if (bind(sock, (struct sockaddr *)&iface->llink, sizeof(struct sockaddr_in6)) == -1) {
-				ERROR("%s: bind: %s", __func__, strerror(errno));
-				goto exit_err_0;
+				ERROR("%s: bind: %s\n", __func__, strerror(errno));
+				goto err_mld_query_msg_free;
 			}
 		}
 	}
 	if (setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_IF, &ifx, sizeof(ifx)) == -1)
-		ERROR("setsockopt");
+		ERROR("setsockopt\n");
 
 	for (int i = 0; i < MLD2_ROBUSTNESS; i++) {
-		if (sendmsg(sock, &msg.msgh, 0) == -1) ERROR("sendmsg");
+		if (sendmsg(sock, &msg.msgh, 0) == -1) ERROR("sendmsg\n");
 	}
-exit_err_0:
+err_mld_query_msg_free:
 	mld_query_msg_free(&msg);
 }
 
@@ -862,25 +959,25 @@ static void mld_address_record(mld_t *mld, unsigned int ifx, mld_addr_rec_t *rec
 {
 	struct in6_addr grp = rec->addr;
 #ifdef DEBUG_ON
-	char strgrp[INET6_ADDRSTRLEN];
-	char ifname[IF_NAMESIZE];
+	char strgrp[INET6_ADDRSTRLEN] = {0};
+	char ifname[IF_NAMESIZE] = {0};
 	if_indextoname(ifx, ifname);
 	inet_ntop(AF_INET6, &grp, strgrp, INET6_ADDRSTRLEN);
 #endif
 
-	DEBUG("%s type=%u", __func__, rec->type);
+	//DEBUG("%s type=%u\n", __func__, rec->type);
 
 	/* XXX: we're ignoring SSM here */
 	switch (rec->type) {
 		case BLOCK_OLD_SOURCES:
 		case MODE_IS_INCLUDE:
 		case CHANGE_TO_INCLUDE_MODE:
-			DEBUG("INCLUDE %s received on %s(%u)", strgrp, ifname, ifx);
+			DEBUG("INCLUDE %s received on %s(%u)\n", strgrp, ifname, ifx);
 			mld_filter_grp_part(mld, ifx, &grp);
 			break;
 		case MODE_IS_EXCLUDE:
 		case CHANGE_TO_EXCLUDE_MODE:
-			DEBUG("EXCLUDE %s received on %s(%u)", strgrp, ifname, ifx);
+			DEBUG("EXCLUDE %s received on %s(%u)\n", strgrp, ifname, ifx);
 			mld_filter_grp_add(mld, ifx, &grp);
 			break;
 	}
@@ -895,19 +992,18 @@ void mld_query_handler(mld_t *mld, unsigned int ifx, struct ip6_hdr *ip6h)
 	struct in6_addr *src = (struct in6_addr *)((char *)ip6h + offsetof(struct ip6_hdr, ip6_src));
 
 	if (!IN6_IS_ADDR_LINKLOCAL(src)) {
-		DEBUG("MLD Query source is not link local - ignoring");
+		DEBUG("MLD Query source is not link local - ignoring\n");
 		return;
 	}
 
 	// TODO ignore if hop limit != 1
-	// TODO ignore if router alert NH header not set
 
 #if DEBUG_ON
-	char straddrgrp[INET6_ADDRSTRLEN];
-	char ifname[IF_NAMESIZE];
+	char straddrgrp[INET6_ADDRSTRLEN] = {0};
+	char ifname[IF_NAMESIZE] = {0};
 	inet_ntop(AF_INET6, src, straddrgrp, INET6_ADDRSTRLEN);
 	if_indextoname(ifx, ifname);
-	DEBUG("processing MLD Query from %s on %s(%u)", straddrgrp, ifname, ifx);
+	DEBUG("processing MLD Query from %s on %s(%u)\n", straddrgrp, ifname, ifx);
 #endif
 	iface = mld_get_iface(mld, ifx);
 	if (!iface) return;
@@ -918,11 +1014,10 @@ void mld_query_handler(mld_t *mld, unsigned int ifx, struct ip6_hdr *ip6h)
 	if (be64toh(srclo64) < be64toh(loclo64)) {
 		/* other node is Querier - set timer */
 		struct timespec now;
-		clock_gettime(CLOCK_REALTIME, &now);
-		if (mld_is_querier(mld, ifx, &now)) {
+		if (!clock_gettime(CLOCK_REALTIME, &now) && mld_is_querier(mld, ifx, &now) == 1) {
 			char ifname[IF_NAMESIZE];
 			if_indextoname(ifx, ifname);
-			DEBUG("Querier detected. Entering Non-Querier state on %s(%u)", ifname, ifx);
+			DEBUG("Querier detected. Entering Non-Querier state on %s(%u)\n", ifname, ifx);
 		}
 		iface->qseen.tv_sec = now.tv_sec;
 	}
@@ -933,7 +1028,7 @@ static void mld2_listen_report(mld_t *mld, unsigned int ifx, struct mld_hdr *mld
 	struct icmp6_hdr *icmpv6 = (struct icmp6_hdr *)mldh;
 	uint16_t recs = ntohs(icmpv6->icmp6_data16[1]);
 	mld_addr_rec_t *mrec = (mld_addr_rec_t *)((char *)mldh + offsetof(struct mld_hdr, mld_addr));
-	DEBUG("processing mld2 listen report with %u records", recs);
+	DEBUG("processing mld2 listen report with %u records\n", recs);
 	while (recs--) mld_address_record(mld, ifx, mrec++);
 }
 
@@ -942,25 +1037,25 @@ static void mld_msg_handler(mld_t *mld, unsigned int ifx, struct ip6_hdr *ip6h,
 	struct in6_addr *mld_addr = (struct in6_addr *)((char *)mldh + offsetof(struct mld_hdr, mld_addr));
 	switch (mldh->mld_type) {
 		case MLD_LISTENER_QUERY:
-			DEBUG("MLD_LISTENER_QUERY received on %u", ifx);
+			DEBUG("MLD_LISTENER_QUERY received on %u\n", ifx);
 			mld_query_handler(mld, ifx, ip6h);
 			break;
 		case MLD2_LISTENER_REPORT:
-			DEBUG("MLD2_LISTENER_REPORT received on %u", ifx);
+			DEBUG("MLD2_LISTENER_REPORT received on %u\n", ifx);
 			mld2_listen_report(mld, ifx, mldh);
 			break;
 		case MLD_LISTENER_REPORT:
-			DEBUG("MLD_LISTENER_REPORT received on %u", ifx);
+			DEBUG("MLD_LISTENER_REPORT received on %u\n", ifx);
 			// TODO: set mldv1 compat mode
 			mld_filter_grp_add(mld, ifx, mld_addr);
 			break;
 		case MLD_LISTENER_DONE:
-			DEBUG("MLD_LISTENER_DONE received on %u", ifx);
+			DEBUG("MLD_LISTENER_DONE received on %u\n", ifx);
 			// TODO: set mldv1 compat mode
 			mld_filter_grp_part(mld, ifx, mld_addr);
 			break;
 		default:
-			DEBUG("unhandled icmpv6 type %u received on %u", mldh->mld_type, ifx);
+			DEBUG("unhandled icmpv6 type %u received on %u\n", mldh->mld_type, ifx);
 	}
 }
 
@@ -970,7 +1065,6 @@ static void mld_msg_handler(mld_t *mld, unsigned int ifx, struct ip6_hdr *ip6h,
 static int mld_listen(mld_t *mld)
 {
 	char buf[32767] = {0};
-	//char ctrl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
 	struct sockaddr_ll sll = {0};
 	struct sock_fprog bpf = {
 		.len = sizeof(BPF_MLD) / sizeof(BPF_MLD[0]),
@@ -984,16 +1078,16 @@ static int mld_listen(mld_t *mld)
 	char ipv6[48] = {0};
 	ssize_t byt;
 	unsigned int ifx;
-	int sock;
+	int rc = -1, sock;
 
 	sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
 	if (sock == -1) return -1;
 	if (setsockopt(sock, SOL_SOCKET, SO_ATTACH_FILTER, &bpf, sizeof(bpf)) == -1) {
-		ERROR("setsockopt(SO_ATTACH_FILTER)");
-		close(sock);
-		return -1;
+		ERROR("setsockopt(SO_ATTACH_FILTER)\n");
+		goto err_close_sock;
 	}
 
+	mld->listen_sock = sock;
 	/* prepare scatter/gather structs */
 	iov[0].iov_base = &ethh;
 	iov[0].iov_len = sizeof ethh;
@@ -1003,25 +1097,25 @@ static int mld_listen(mld_t *mld)
 	iov[2].iov_len = sizeof buf;
 	msg.msg_name = &sll;
 	msg.msg_namelen = sizeof sll;
-	//msg.msg_control = ctrl;
-	//msg.msg_controllen = sizeof ctrl;
 	msg.msg_iov = iov;
 	msg.msg_iovlen = sizeof iov / sizeof iov[0];
 	msg.msg_flags = 0;
 	while (1) {
 		byt = recvmsg(sock, &msg, 0);
-		mldh = (struct mld_hdr *)(buf);
-		ip6h = (struct ip6_hdr *)&ipv6;
 		if (byt == -1) {
 			ERROR("read/recvmsg");
-			break;
+			goto err_close_sock;
 		}
+		mldh = (struct mld_hdr *)(buf);
+		ip6h = (struct ip6_hdr *)&ipv6;
 		ifx = (unsigned int)sll.sll_ifindex;
-		DEBUG("received type %u on %u (%zi bytes)", mldh->mld_icmp6_hdr.icmp6_type, ifx, byt);
+		DEBUG("received type %u on %u (%zi bytes)\n", mldh->mld_icmp6_hdr.icmp6_type, ifx, byt);
 		if (ifx) mld_msg_handler(mld, ifx, ip6h, mldh);
 	}
+	rc = 0;
+err_close_sock:
 	close(sock);
-	return 0;
+	return rc;
 }
 #endif
 
@@ -1048,8 +1142,7 @@ static int mld_listen(mld_t *mld)
 	ssize_t byt;
 	unsigned int blen = 32767;
 	unsigned int ifx;
-	int nfds = 0;
-	int rc;
+	int err = 0, nfds = 0, rc = 0;
 
 	for (mld_iface_t *iface = mld->iface; iface; iface = iface->next) nfds++;
 	if (!nfds) return -1;
@@ -1062,6 +1155,7 @@ static int mld_listen(mld_t *mld)
 
 	/* create array of pollfd structures */
 	fds = calloc(nfds, sizeof (struct pollfd));
+	if (!fds) { err = errno; rc = -1; goto err_free_buf; }
 	mld_iface_t *iface = mld->iface;
 	for (int i = 0; i < nfds; i++) {
 		fds[i].fd = iface->bpf;
@@ -1069,8 +1163,9 @@ static int mld_listen(mld_t *mld)
 		iface = iface->next;
 	}
 
-	while (1) {
+	while (nfds > 0) {
 		rc = poll(fds, (nfds_t) nfds, 1000);
+		pthread_testcancel();
 		if (rc > 0) {
 			for (int i = 0; i < nfds; i++) {
 				if (fds[i].revents & POLLIN) {
@@ -1084,47 +1179,49 @@ static int mld_listen(mld_t *mld)
 						ifx = bpffd2ifx(mld, fds[i].fd);
 						if (ifx) mld_msg_handler(mld, ifx, ip6h, mldh);
 					}
+				} else if (fds[i].revents & (POLLHUP | POLLERR)) {
+					/* socket closed or other error, stop checking it */
+					nfds--;
+					if (i != nfds) fds[i] = fds[nfds];
+					i--;
 				}
 			}
 		}
+		if (rc < 0 && errno != EINTR) {
+			err = errno;
+			break;
+		}
 	}
 	free(fds);
+	if (nfds == 0 && err == 0) {
+		err = EIO;
+		rc = -1;
+	}
+err_free_buf:
 	free(buf);
-	return 0;
+	if (err) errno = err;
+	return rc;
 }
 #endif
 
 static void *mld_thread_listen(void *arg)
 {
 	mld_t *mld = (mld_t *)arg;
-	int sock;
-	const int opt = 1;
-	sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
-	if (sock == -1) {
-		ERROR("socket(): %s", strerror(errno));
-		goto exit_0;
-	}
-	if (setsockopt(sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &opt, sizeof(opt))) {
-		ERROR("%s() setsockopt(IPV6_RECVPKTINFO): %s", __func__, strerror(errno));
-		goto exit_1;
-	}
-	mld->sock = sock;
-	sem_post(&mld->sem_mld);   /* release semaphore - we're good to go */
-	sem_wait(&mld->sem_state); /* wait until state thread is ready */
+	/* release semaphore - we're good to go */
+	if (sem_post(&mld->sem_mld) == -1) goto err_close_sock;
+	/* wait until state thread is ready */
+	if (sem_wait(&mld->sem_state) == -1) goto err_close_sock;
 	mld_listen(mld);
 	return NULL;
-exit_1:
-	close(sock);
-exit_0:
+err_close_sock:
+	close(mld->sock);
 	return NULL;
 }
 
 #if HAVE_NETLINK
-static void mld_handle_netlink_msg(mld_t *mld, struct sockaddr_nl *nl, struct nlmsghdr *msg)
+static void mld_handle_netlink_msg(mld_t *mld, struct nlmsghdr *msg)
 {
-	(void)nl; /* unused - remove? */
 	struct ifinfomsg *ifi = NLMSG_DATA(msg);
-	//struct ifaddrmsg *ifa = NLMSG_DATA(msg);
 	switch (msg->nlmsg_type) {
 		case RTM_NEWLINK:
 			if (ifi->ifi_flags & IFF_UP) {
@@ -1135,11 +1232,11 @@ static void mld_handle_netlink_msg(mld_t *mld, struct sockaddr_nl *nl, struct nl
 			}
 			break;
 		case RTM_DELLINK:
-			DEBUG("RTM_DELLINK");
+			DEBUG("RTM_DELLINK\n");
 			mld_filter_ifx_del(mld, ifi->ifi_index);
 			break;
 		default:
-			DEBUG("unknown netlink event type: %i", msg->nlmsg_type);
+			DEBUG("unknown netlink event type: %i\n", msg->nlmsg_type);
 			break;
 	}
 }
@@ -1153,7 +1250,8 @@ static void *mld_thread_netlink(void *arg)
 	struct nlmsghdr *nh;
 	struct sockaddr_nl sa;
 	struct msghdr msg = { &sa, sizeof(sa), &iov, 1, NULL, 0, 0 };
-	int fd, len;
+	ssize_t len;
+	int fd;
 
 	memset(&sa, 0, sizeof(sa));
 	sa.nl_family = AF_NETLINK;
@@ -1164,13 +1262,15 @@ static void *mld_thread_netlink(void *arg)
 		return NULL;
 	}
 
+	sem_post(&mld->sem_netlink);
 	while (1) {
 		len = recvmsg(fd, &msg, 0);
+		if (len == -1) continue;
 		for (nh = (struct nlmsghdr *)buf; NLMSG_OK(nh, (size_t)len); nh = NLMSG_NEXT(nh, len)) {
 			/* The end of multipart message */
 			if (nh->nlmsg_type == NLMSG_DONE) continue;
 			if (nh->nlmsg_type == NLMSG_ERROR) ERROR("NLMSG_ERROR\n");
-			else mld_handle_netlink_msg(mld, &sa, nh);
+			else mld_handle_netlink_msg(mld, nh);
 		}
 	}
 
@@ -1186,7 +1286,7 @@ static void *mld_thread_timer(void *arg)
 		/* we use a semaphore to ensure we only queue a maximum of one
 		 * job at a time, even when STATE thread is busy */
 		if (sem_trywait(&mld->sem_state) == -1) continue;
-		job_push_new(mld->q[MLD_STATE], &mld_state_timer_check, mld, sizeof(*mld), &free, 0);
+		job_push_new(mld->q[MLD_STATE], &mld_state_timer_check, mld, sizeof(*mld), NULL, 0);
 	}
 	return arg;
 }
@@ -1195,55 +1295,97 @@ mld_t *mld_init(int flags)
 {
 	(void)flags; /* TODO implement flags */
 	mld_t *mld;
+	int err = 0;
 	if (!(mld = malloc(sizeof(mld_t)))) return NULL;
 	memset(mld, 0, sizeof(mld_t));
-	sem_init(&mld->sem_mld, 0, 0);
-	sem_init(&mld->sem_state, 0, 0);
+	if (sem_init(&mld->sem_mld, 0, 0) == -1) {
+		err = errno; goto err_free_mld;
+	}
+	if (sem_init(&mld->sem_state, 0, 0) == -1) {
+		err = errno; goto err_sem_mld_destroy;
+	}
+#if HAVE_NETLINK
+	if (sem_init(&mld->sem_netlink, 0, 0) == -1) {
+		err = errno; goto err_sem_state_destroy;
+	}
+#endif
+	mld->listen_sock = -1;
 	return mld;
+#if HAVE_NETLINK
+err_sem_state_destroy:
+	sem_destroy(&mld->sem_state);
+#endif
+err_sem_mld_destroy:
+	sem_destroy(&mld->sem_mld);
+err_free_mld:
+	free(mld);
+	if (err) errno = err;
+	return NULL;
 }
 
 void mld_stop(mld_t *mld)
 {
+#ifdef __linux__
+	if (mld->listen_sock >= 0) {
+		close(mld->listen_sock);
+		mld->listen_sock = -1;
+	}
+#else
+	for (mld_iface_t *iface = mld->iface; iface; iface = iface->next) {
+		if (iface->bpf >= 0) {
+			close(iface->bpf);
+			iface->bpf = -1;
+		}
+	}
+#endif
 	for (int i = 0; i < MLD_THREADS; i++) if (mld->q[i]) job_queue_destroy(mld->q[i]);
 }
 
 mld_t *mld_start(mld_t *mld)
 {
 	struct timespec ts;
+	const int opt = 1;
+	int err = 0;
 	if (!mld) mld = mld_init(0);
 	if (!mld) return NULL;
 	for (int i = 0; i < MLD_THREADS; i++) {
 		mld->q[i] = job_queue_create(1);
-		assert(mld->q[i]);
+		if (!mld->q[i]) goto err_mld_stop;
 	}
 #if HAVE_NETLINK
 	mld->job[MLD_NETLINK] = job_push_new(mld->q[MLD_NETLINK],
-			&mld_thread_netlink, mld, sizeof(*mld), &free, 0);
+			&mld_thread_netlink, mld, sizeof(*mld), NULL, 0);
+	if (!mld->job[MLD_NETLINK]) goto err_mld_stop;
+	if (sem_wait(&mld->sem_netlink)) goto err_mld_stop; /* make sure netlink is ready */
 #endif
 
 	/* start listen thread and wait for it to initialize */
+	mld->sock = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+	if (mld->sock == -1) goto err_mld_stop;
+	if (setsockopt(mld->sock, IPPROTO_IPV6, IPV6_RECVPKTINFO, &opt, sizeof(opt)))
+		goto err_mld_stop;
 	mld->job[MLD_LISTEN] = job_push_new(mld->q[MLD_LISTEN],
-			&mld_thread_listen, mld, sizeof(*mld), &free, 0);
-#if MLD_LISTEN_THREAD_STARTUP
-	if (clock_gettime(CLOCK_REALTIME, &ts) == -1) {
-		goto exit_err_0;
-	}
+			&mld_thread_listen, mld, sizeof(*mld), NULL, 0);
+	if (!mld->job[MLD_LISTEN]) goto err_mld_stop;
+	if (clock_gettime(CLOCK_REALTIME, &ts) == -1) goto err_mld_stop;
 	ts.tv_nsec += MLD_LISTEN_THREAD_STARTUP;
-	sem_timedwait(&mld->sem_mld, &ts);
-#endif
-
+	if (ts.tv_nsec > 999999999L) {
+		ts.tv_nsec -= 1000000000L;
+		ts.tv_sec++;
+	}
+	if (sem_timedwait(&mld->sem_mld, &ts) == -1) goto err_mld_stop;
 	/* push existing interfaces to STATE thread */
-	job_push_new(mld->q[MLD_STATE], &mld_state_ifx_add_all, mld, sizeof(*mld), &free, JOB_WAIT);
-	sem_post(&mld->sem_state); /* let mld_thread_listen() proceed */
-	sem_post(&mld->sem_state); /* first mld_thread_timer() run */
-
+	mld_state_ifx_add_all(mld);
+	/* let mld_thread_listen() proceed */
+	if (sem_post(&mld->sem_state) == -1) goto err_mld_stop;
+	/* first mld_thread_timer() run */
+	if (sem_post(&mld->sem_state) == -1) goto err_mld_stop;
 	/* start timer thread */
-	job_push_new(mld->q[MLD_TIMER], &mld_thread_timer, mld, sizeof(*mld), &free, 0);
-
-	return mld;
-#if MLD_LISTEN_THREAD_STARTUP
-exit_err_0:
-	free(mld);
+	if (job_push_new(mld->q[MLD_TIMER], &mld_thread_timer, mld, sizeof(*mld), NULL, 0))
+		return mld;
+err_mld_stop:
+	err = errno;
+	mld_stop(mld);
+	errno = err;
 	return NULL;
-#endif
 }
