@@ -1,20 +1,20 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2017-2022 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2017-2023 Brett Sheffield <bacs@librecast.net> */
 
 #define _GNU_SOURCE
-#include "librecast_pvt.h"
+#include <librecast_pvt.h>
 #include <librecast/net.h>
 #ifdef HAVE_LIBLCRQ
 #include <lcrq.h>
 #endif
+#ifdef HAVE_MLD
+#include <mld.h>
+#endif
 #include "hash.h"
+#include <librecast/sync.h>
+#include <librecast/mdex.h>
 #include <arpa/inet.h>
 #include <assert.h>
-#ifdef HAVE_ENDIAN_H
-#include <endian.h>
-#elif HAVE_SYS_ENDIAN_H
-#include <sys/endian.h>
-#endif
 #include <fcntl.h>
 #include <ifaddrs.h>
 #include <net/if.h>
@@ -23,15 +23,12 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/param.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <poll.h>
 
-uint32_t ctx_id = 0;
-uint32_t sock_id = 0;
-uint32_t chan_id = 0;
-
 lc_ctx_t *ctx_list = NULL;
 
 static void lc_op_data_handler(lc_socket_call_t *sc, lc_message_t *msg);
@@ -43,6 +40,8 @@ static ssize_t lc_socket_decoded(lc_socket_t *sock, struct msghdr *msg, ssize_t
 #endif
 
 int (*lc_msg_logger)(lc_channel_t *, lc_message_t *, void *logdb) = NULL;
+static void socket_close(lc_socket_t *sock);
+static void channel_free(lc_channel_t * chan);
 
 void (*lc_op_handler[LC_OP_MAX])(lc_socket_call_t *, lc_message_t *) = {
 	lc_op_data_handler,
@@ -61,26 +60,18 @@ int lc_getrandom(void *buf, size_t buflen)
 	return err;
 }
 
-uint32_t lc_ctx_get_id(lc_ctx_t *ctx)
-{
-	return (ctx) ? ctx->id : 0;
-}
-
-uint32_t lc_socket_get_id(lc_socket_t *sock)
-{
-	return (sock) ? sock->id : 0;
-}
-
-uint32_t lc_channel_get_id(lc_channel_t *chan)
-{
-	return (chan) ? chan->id : 0;
-}
-
 lc_ctx_t *lc_channel_ctx(lc_channel_t *chan)
 {
 	return chan->ctx;
 }
 
+#ifdef HAVE_LIBLCRQ
+rq_t *lc_channel_rq(lc_channel_t *chan)
+{
+	return chan->rq;
+}
+#endif
+
 lc_socket_t *lc_channel_socket(lc_channel_t *chan)
 {
 	return chan->sock;
@@ -106,6 +97,23 @@ int lc_channel_socket_raw(lc_channel_t *chan)
 	return chan->sock->sock;
 }
 
+void lc_ctx_ifx(lc_ctx_t *ctx, unsigned int ifx)
+{
+	ctx->ifx = ifx;
+}
+
+void lc_ctx_ratelimit(lc_ctx_t *ctx, size_t bps_out, size_t bps_in)
+{
+	ctx->bps_out = bps_out;
+	ctx->bps_in = bps_in;
+}
+
+void lc_channel_ratelimit(lc_channel_t *chan, size_t bps_out, size_t bps_in)
+{
+	chan->bps_out = bps_out;
+	chan->bps_in = bps_in;
+}
+
 int lc_socket_raw(lc_socket_t *sock)
 {
 	return sock->sock;
@@ -157,7 +165,7 @@ int lc_msg_set(lc_message_t *msg, lc_msg_attr_t attr, void *value)
 int lc_msg_id(lc_message_t *msg, unsigned char *id, size_t len)
 {
 #ifndef HASH_TYPE
-	errno = ENOTSUP; return -1;
+	return (errno = ENOTSUP), -1;
 #else
 	hash_state state;
 	hash_init(&state, NULL, 0, len);
@@ -228,12 +236,41 @@ int lc_msg_init_size(lc_message_t *msg, size_t len)
 	return 0;
 }
 
+int lc_ctx_coding_set(lc_ctx_t *ctx, int coding)
+{
+	return (ctx->coding = coding);
+}
+
 int lc_channel_coding_set(lc_channel_t *chan, int coding)
 {
 	if (chan->sock) chan->sock->ccoding = coding;
 	return (chan->coding = coding);
 }
 
+static int lc_ctx_keygetorset(lc_ctx_t *ctx, lc_key_t *key, int type, int set)
+{
+	lc_key_t *k;
+	switch (type) {
+		case LC_CODE_SYMM:
+			k = &ctx->key;
+			break;
+		case LC_CODE_PUBK:
+			k = &ctx->pek;
+			break;
+		default:
+			return (errno = EINVAL), -1;
+	}
+	if (set) {
+		k->key = key->key;
+		k->keylen = key->keylen;
+	}
+	else {
+		key->key = k->key;
+		key->keylen = k->keylen;
+	}
+	return 0;
+}
+
 static int lc_channel_keygetorset(lc_channel_t *chan, lc_key_t *key, int type, int set)
 {
 	lc_key_t *k;
@@ -245,13 +282,9 @@ static int lc_channel_keygetorset(lc_channel_t *chan, lc_key_t *key, int type, i
 			k = &chan->pek;
 			break;
 		default:
-			return -1;
+			return (errno = EINVAL), -1;
 	}
 	if (set) {
-		/* before storing keys mlock channel memory */
-#ifdef HAVE_LIBSODIUM
-		if (sodium_mlock(chan, sizeof *chan)) return -1;
-#endif
 		k->key = key->key;
 		k->keylen = key->keylen;
 	}
@@ -262,6 +295,28 @@ static int lc_channel_keygetorset(lc_channel_t *chan, lc_key_t *key, int type, i
 	return 0;
 }
 
+int lc_ctx_getkey(lc_ctx_t *ctx, lc_key_t *key, int type)
+{
+	return lc_ctx_keygetorset(ctx, key, type, 0);
+}
+
+int lc_ctx_setkey(lc_ctx_t *ctx, lc_key_t *key, int type)
+{
+	return lc_ctx_keygetorset(ctx, key, type, 1);
+}
+
+int lc_ctx_set_sym_key(lc_ctx_t *ctx, unsigned char *key, size_t len)
+{
+	lc_key_t k = { .key = key, .keylen = len };
+	return lc_ctx_setkey(ctx, &k, LC_CODE_SYMM);
+}
+
+int lc_ctx_set_pub_key(lc_ctx_t *ctx, unsigned char *key, size_t len)
+{
+	lc_key_t k = { .key = key, .keylen = len };
+	return lc_ctx_setkey(ctx, &k, LC_CODE_PUBK);
+}
+
 int lc_channel_getkey(lc_channel_t *chan, lc_key_t *key, int type)
 {
 	return lc_channel_keygetorset(chan, key, type, 0);
@@ -298,8 +353,10 @@ static void lc_close_nack_logger(lc_channel_logging_t *logging)
 }
 
 static void free_gap_detect(lc_channel_gap_detect_t * gap_detect) {
-	lc_socket_close(gap_detect->sock_nack);
-	lc_channel_free(gap_detect->chan_nack);
+	socket_close(gap_detect->sock_nack);
+	free(gap_detect->sock_nack);
+	channel_free(gap_detect->chan_nack);
+	free(gap_detect->chan_nack);
 	free(gap_detect);
 }
 
@@ -390,21 +447,77 @@ int lc_channel_check_seqno(lc_channel_t *chan, lc_seq_t seq)
 	else return 0;
 }
 
-void lc_channel_free(lc_channel_t * chan)
+static void channel_free(lc_channel_t * chan)
 {
-	if (!chan) return;
 	for (lc_channel_t *p = chan->ctx->chan_list, *prev = NULL; p; p = p->next) {
-		if (p->id == chan->id) {
+		if (p == chan) {
 			if (prev) prev->next = p->next;
 			else chan->ctx->chan_list = p->next;
+			break;
 		}
 		prev = p;
 	}
+#ifdef HAVE_LIBLCRQ
+	if (chan->rq) rq_free(chan->rq);
+#endif
 	if (chan->logging) lc_close_nack_logger(chan->logging);
 	if (chan->gap_detect) free_gap_detect(chan->gap_detect);
-	free(chan);
 }
 
+void lc_channel_free(lc_channel_t * chan)
+{
+	int err = errno;
+	if (!chan) return;
+	if (!pthread_mutex_lock(&chan->ctx->mtx)) {
+		channel_free(chan);
+		pthread_mutex_unlock(&chan->ctx->mtx);
+	}
+	free(chan);
+	errno = err;
+}
+
+static void chan_ratelimit(lc_channel_t *chan)
+{
+	uint64_t nss, nse;
+	double s, bps;
+	struct timespec now = {0};
+	nss = chan->byt_reset.tv_nsec + chan->byt_reset.tv_sec * NANO;
+	while (1) {
+		clock_gettime(CLOCK_TAI, &now);
+		nse = now.tv_nsec + now.tv_sec * NANO;
+		s = (double)(nse - nss) / NANO;
+		bps = chan->byt_out * 8 / s;
+		if (bps < chan->bps_out) break;
+		usleep(100);
+	}
+}
+
+inline static ssize_t chan_sendmsg(lc_channel_t *chan, const struct msghdr *msg, int flags)
+{
+	ssize_t rc = 0;
+	if (chan->bps_out) chan_ratelimit(chan);
+	if (flags & MSG_DROP) {
+		for (int i = 0; i < (int)msg->msg_iovlen; i++) rc += msg->msg_iov[i].iov_len;
+	}
+	else {
+		rc = sendmsg(chan->sock->sock, msg, flags);
+		if (rc > 0) chan->byt_out += rc;
+	}
+	return rc;
+}
+
+inline static ssize_t chan_sendto(lc_channel_t *chan, const void *buf, size_t len, int flags,
+	const struct sockaddr *dest_addr, socklen_t addrlen)
+{
+	ssize_t rc;
+	if (chan->bps_out) chan_ratelimit(chan);
+	if (flags & MSG_DROP) return len;
+	else rc = sendto(chan->sock->sock, buf, len, flags, dest_addr, addrlen);
+	if (rc > 0) chan->byt_out += rc;
+	return rc;
+}
+
+
 #ifdef HAVE_LIBSODIUM
 static ssize_t lc_channel_sendmsg_encoded(lc_channel_t *chan, struct msghdr *msg, int flags)
 {
@@ -436,7 +549,7 @@ static ssize_t lc_channel_sendmsg_encoded(lc_channel_t *chan, struct msghdr *msg
 	msg->msg_iov = iov;
 	msg->msg_iovlen = 2;
 
-	byt = sendmsg(chan->sock->sock, msg, flags);
+	byt = chan_sendmsg(chan, msg, flags);
 
 	free(ebuf);
 
@@ -448,58 +561,68 @@ ssize_t lc_channel_sendmsg(lc_channel_t *chan, struct msghdr *msg, int flags)
 {
 	msg->msg_name = (struct sockaddr *)&chan->sa;
 	msg->msg_namelen = sizeof(struct sockaddr_in6);
-	if ((chan->coding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (chan->coding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 		return lc_channel_sendmsg_encoded(chan, msg, flags);
 #else
-		errno = ENOTSUP; return -1;
+		return (errno = ENOTSUP), -1;
 #endif
 	}
-	return sendmsg(chan->sock->sock, msg, flags);
+	return chan_sendmsg(chan, msg, flags);
 }
 
 static ssize_t lc_channel_send_coded(lc_channel_t *chan, const void *buf, size_t len, int flags)
 {
 #ifdef HAVE_LIBLCRQ
-	if ((chan->coding & LC_CODE_FEC_RQ) == LC_CODE_FEC_RQ) {
-		static rq_pid_t pid = 0;
+	if (chan->coding & LC_CODE_FEC_RQ) {
+		uint32_t esi = 0;
+		uint8_t sbn = 0;
+		rq_pid_t pid = 0;
 		if (buf) {
 			/* new data, encode */
-			pid = 0;
+			chan->pid = pid;
 			if (!chan->rq) chan->rq = rq_init(len, chan->T);
 			if (!chan->rq) return -1;
-			int rc = rq_encode(chan->rq, (uint8_t *)buf, len);
-			if (rc) return -1;
+			if (rq_encode(chan->rq, (uint8_t *)buf, len)) return -1;
 		}
 		else {
-			uint8_t sbn = pid >> 24;
-			uint32_t esi = ntohl(pid << 8);
+			pid = chan->pid;
+			sbn = rq_pid2sbn(pid);
+			esi = rq_pid2esi(pid);
 			if (!chan->rq) return -1; /* no rq object */
-			if ((uint32_t)(rq_KP(chan->rq) + RQ_OVERHEAD) == ++esi) {
+			if ((uint32_t)(rq_KP(chan->rq) + RQ_OVERHEAD * 2) == ++esi) {
 				esi = 0;
 				if (rq_Z(chan->rq) == ++sbn) sbn = 0;
 			}
-			pid = htonl(esi) >> 8;
-			pid |= (sbn << 24);
+			pid = rq_pidset(sbn, esi);
+			chan->pid = pid;
 		}
 		/* send next symbol */
-		struct iovec iov[2];
-		struct msghdr msg = {
-			.msg_iov = iov,
-			.msg_iovlen = sizeof iov / sizeof iov[0]
-		};
+		struct iovec iov[4];
+		struct msghdr msg = { .msg_iov = iov, };
 		uint8_t sym[chan->T];
-		const int rqf = ((chan->coding & LC_CODE_FEC_RAND) == LC_CODE_FEC_RAND) ? RQ_RAND : 0;
-		iov[0].iov_base = &pid;
-		iov[0].iov_len = sizeof pid;
-		iov[1].iov_base = sym;
-		iov[1].iov_len = sizeof sym;
+		const int rqf = ((chan->coding & LC_CODE_FEC_RAND) || esi >= rq_K(chan->rq)) ? RQ_RAND : 0;
+		rq_oti_t oti = 0;
+		rq_scheme_t scheme = 0;
+		int v = 0;
+		if (chan->coding & LC_CODE_FEC_OTI) {
+			rq_oti(chan->rq, &oti, &scheme);
+			iov[v].iov_base = &oti;
+			iov[v++].iov_len = sizeof oti;
+			iov[v].iov_base = &scheme;
+			iov[v++].iov_len = sizeof scheme;
+		}
+		iov[v].iov_base = &pid;
+		iov[v++].iov_len = sizeof pid;
+		iov[v].iov_base = sym;
+		iov[v++].iov_len = sizeof sym;
+		msg.msg_iovlen = v;
 		rq_symbol(chan->rq, &pid, sym, rqf);
 		return lc_channel_sendmsg(chan, &msg, flags);
 	}
 #endif
 #ifdef HAVE_LIBSODIUM
-	const size_t elen = ((chan->coding & LC_CODE_SYMM) != LC_CODE_SYMM) ? len :
+	const size_t elen = (!(chan->coding & LC_CODE_SYMM)) ? len :
 		len + crypto_secretbox_MACBYTES + crypto_secretbox_NONCEBYTES;
 	unsigned char ebuf[elen];
 	memset(ebuf, 0, elen);
@@ -508,53 +631,63 @@ static ssize_t lc_channel_send_coded(lc_channel_t *chan, const void *buf, size_t
 #endif
 	unsigned char *sbuf = (unsigned char *)buf;
 
-	if (buf && (chan->coding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (buf && (chan->coding & LC_CODE_SYMM)) {
 #ifdef HAVE_LIBSODIUM
 		unsigned char *nonce = ebuf + len + crypto_secretbox_MACBYTES;
 		randombytes_buf(nonce, crypto_secretbox_MACBYTES);
 		crypto_secretbox_easy(ebuf, (unsigned char *)buf, len, nonce, chan->key.key);
 		sbuf = ebuf;
 #else
-		errno = ENOTSUP; return -1;
+		return (errno = ENOTSUP), -1;
 #endif
 	}
-	return sendto(chan->sock->sock, sbuf, elen, flags,
-		(struct sockaddr *)&chan->sa, sizeof(struct sockaddr_in6));
+	return chan_sendto(chan, sbuf, elen, flags, (struct sockaddr *)&chan->sa, sizeof(struct sockaddr_in6));
 }
 
 ssize_t lc_channel_send(lc_channel_t *chan, const void *buf, size_t len, int flags)
 {
+	ssize_t rc;
 	if (chan->coding) return lc_channel_send_coded(chan, buf, len, flags);
-	return sendto(chan->sock->sock, buf, len, flags,
-		(struct sockaddr *)&chan->sa, sizeof(struct sockaddr_in6));
+	rc = chan_sendto(chan, buf, len, flags, (struct sockaddr *)&chan->sa, sizeof(struct sockaddr_in6));
+	return rc;
 }
 
 ssize_t lc_socket_sendmsg(lc_socket_t *sock, struct msghdr *msg, int flags)
 {
-	ssize_t bytes = 0, rc;
-	for (lc_channel_t *chan = sock->ctx->chan_list; chan; chan = chan->next) {
-		if (chan->sock == sock) {
-			if ((rc = lc_channel_sendmsg(chan, msg, flags)) > 0) {
-				bytes += rc;
+	ssize_t bytes = 0, rc = -1;
+	if (!pthread_mutex_lock(&sock->ctx->mtx)) {
+		for (lc_channel_t *chan = sock->ctx->chan_list; chan; chan = chan->next) {
+			pthread_mutex_unlock(&sock->ctx->mtx);
+			if (chan->sock == sock) {
+				if ((rc = lc_channel_sendmsg(chan, msg, flags)) > 0) {
+					bytes += rc;
+				}
+				else return -1;
 			}
-			else return -1;
+			if (pthread_mutex_lock(&sock->ctx->mtx)) return -1;
 		}
+		pthread_mutex_unlock(&sock->ctx->mtx);
 	}
-	return bytes;
+	return (rc > 0) ? bytes : -1;
 }
 
 ssize_t lc_socket_send(lc_socket_t *sock, const void *buf, size_t len, int flags)
 {
-	ssize_t bytes = 0, rc;
-	for (lc_channel_t *chan = sock->ctx->chan_list; chan; chan = chan->next) {
-		if (chan->sock == sock) {
-			if ((rc = lc_channel_send(chan, buf, len, flags)) > 0) {
-				bytes += rc;
+	ssize_t bytes = 0, rc = -1;
+	if (!pthread_mutex_lock(&sock->ctx->mtx)) {
+		for (lc_channel_t *chan = sock->ctx->chan_list; chan; chan = chan->next) {
+			pthread_mutex_unlock(&sock->ctx->mtx);
+			if (chan->sock == sock) {
+				if ((rc = lc_channel_send(chan, buf, len, flags)) > 0) {
+					bytes += rc;
+				}
+				else return -1;
 			}
-			else return -1;
+			if (pthread_mutex_lock(&sock->ctx->mtx)) return -1;
 		}
+		pthread_mutex_unlock(&sock->ctx->mtx);
 	}
-	return bytes;
+	return (rc > 0) ? bytes : -1;
 }
 
 int lc_channel_nack_handler_thr(lc_channel_t *chan, int n_seconds, void *(*nack_thread)(void *))
@@ -683,6 +816,7 @@ int lc_channel_nack_add_log(lc_channel_t *chan, const void *buf, size_t len, lc_
 
 ssize_t lc_msg_sendto(int sock, const void *buf, size_t len, struct sockaddr_in6 *sa, int flags)
 {
+	if (flags & MSG_DROP) return len;
 	return sendto(sock, buf, len, flags, (struct sockaddr *)sa, sizeof(struct sockaddr_in6));
 }
 
@@ -716,12 +850,12 @@ ssize_t lc_msg_send(lc_channel_t *chan, lc_message_t *msg)
 	head->len = htobe64(msg->len);
 	head->op = msg->op;
 	len = msg->len;
-	if ((chan->coding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (chan->coding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 		ohead = crypto_secretbox_MACBYTES + crypto_secretbox_NONCEBYTES;
 #else
 		free(head);
-		errno = ENOTSUP; return -1;
+		return (errno = ENOTSUP), -1;
 #endif
 	}
 	buf = calloc(1, sizeof(lc_message_head_t) + len + ohead);
@@ -733,7 +867,7 @@ ssize_t lc_msg_send(lc_channel_t *chan, lc_message_t *msg)
 	memcpy(buf + sizeof(lc_message_head_t), msg->data, len);
 	len += sizeof(lc_message_head_t);
 
-	if ((chan->coding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (chan->coding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 		unsigned char *nonce;
 		nonce = (unsigned char *)buf + len + crypto_secretbox_MACBYTES;
@@ -747,7 +881,7 @@ ssize_t lc_msg_send(lc_channel_t *chan, lc_message_t *msg)
 		/* chan->seq could have been changed by another thread in the meantime,
 		 * so use head->seq which we know hasn't */
 		lc_channel_logmsg(chan->logging, buf, len + ohead, be64toh(head->seq));
-	bytes = lc_msg_sendto(chan->sock->sock, buf, len + ohead, sa, 0);
+	bytes = chan_sendto(chan, buf, len + ohead, 0, (struct sockaddr *)sa, sizeof(struct sockaddr_in6));
 	if (bytes == -1) err = errno;
 
 	free(head);
@@ -770,11 +904,15 @@ static int lc_socket_group_joined(lc_socket_t *sock, struct in6_addr *grp)
 
 lc_channel_t *lc_channel_by_address(lc_ctx_t *lctx, struct in6_addr *addr)
 {
-	for (lc_channel_t *p = lctx->chan_list; p; p = p->next) {
-		if (!memcmp(addr,& p->sa.sin6_addr, sizeof(struct in6_addr)))
-			return p;
+	lc_channel_t *p = NULL;
+	if (!pthread_mutex_lock(&lctx->mtx)) {
+		for (p = lctx->chan_list; p; p = p->next) {
+			if (!memcmp(addr,& p->sa.sin6_addr, sizeof(struct in6_addr)))
+				break;
+		}
+		pthread_mutex_unlock(&lctx->mtx);
 	}
-	return NULL;
+	return p;
 }
 
 ssize_t lc_msg_recv(lc_socket_t *sock, lc_message_t *msg)
@@ -819,11 +957,11 @@ recv_again:
 			break;
 		}
 	}
-	if ((sock->ccoding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (sock->ccoding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 		zi = lc_socket_decoded(sock, &msgh, zi, &pi);
 #else
-		errno = ENOTSUP; return -1;
+		return (errno = ENOTSUP), -1;
 #endif
 	}
 	if (lc_msg_init_size(msg, (size_t)zi - sizeof(lc_message_head_t)) == -1)
@@ -903,7 +1041,18 @@ static ssize_t lc_socket_recvmsg_if(lc_socket_t *sock, struct msghdr *msg, int f
 		for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
 			if (cmsg->cmsg_type == IPV6_PKTINFO) {
 				memcpy(pi, CMSG_DATA(cmsg), sizeof(struct in6_pktinfo));
+#ifndef IPV6_MULTICAST_ALL
+				if (!lc_socket_group_joined(sock, &pi->ipi6_addr)) {
+					if (flags & MSG_PEEK) {
+						/* read & drop packet */
+						int beijing = flags^MSG_PEEK;
+						recvmsg(sock->sock, msg, beijing);
+					}
+				}
+				else
+#endif
 				if (!sock->ifx || sock->ifx == pi->ipi6_ifindex) return bytes;
+				else break;
 			}
 		}
 	}
@@ -919,14 +1068,15 @@ static ssize_t lc_socket_decoded(lc_socket_t *sock, struct msghdr *msg, ssize_t
 	unsigned char *base = (unsigned char *)msg->msg_iov[0].iov_base;
 	unsigned char *nonce = base + elen;
 
-	if (bytes < crypto_secretbox_NONCEBYTES + crypto_secretbox_MACBYTES + 1)
+	if (bytes < crypto_secretbox_NONCEBYTES + crypto_secretbox_MACBYTES + 1) {
 		return -1;
+	}
 
 	chan = lc_channel_by_address(sock->ctx, &pi->ipi6_addr);
 	if (!chan) return -1;
 
 	if (crypto_secretbox_open_easy(base, base, elen, nonce, chan->key.key))
-		return -1;
+		return (errno = EBADMSG), -1;
 
 	return bytes - crypto_secretbox_MACBYTES - crypto_secretbox_NONCEBYTES;
 }
@@ -942,6 +1092,7 @@ static ssize_t lc_socket_recvmsg_encoded(lc_socket_t *sock, struct msghdr *msg,
 	msg->msg_control = ctl;
 	msg->msg_controllen = sizeof ctl;
 	size_t sz = crypto_secretbox_MACBYTES;
+	int err = 0;
 
 	for (int i = 0; i < (int)msg->msg_iovlen; i++) {
 		sz += msg->msg_iov[i].iov_len;
@@ -956,19 +1107,19 @@ static ssize_t lc_socket_recvmsg_encoded(lc_socket_t *sock, struct msghdr *msg,
 	bytes = lc_socket_recvmsg_if(sock, &emsg, flags, &pi);
 	if (bytes < crypto_secretbox_NONCEBYTES + crypto_secretbox_MACBYTES + 1) {
 		bytes = -1;
-		goto err_0;
+		goto err_free_buf;
 	}
 	lc_channel_t *chan;
 	chan = lc_channel_by_address(sock->ctx, &pi.ipi6_addr);
 	if (!chan) {
 		bytes = -1;
-		goto err_0;
+		goto err_free_buf;
 	}
 
 	nonce = buf + sz;
 	if (crypto_secretbox_open_easy(buf, buf, sz, nonce, chan->key.key)) {
-		bytes = -1;
-		goto err_0;
+		bytes = -1; err = EBADMSG;
+		goto err_free_buf;
 	}
 	/* scatter parts */
 	ptr = buf;
@@ -976,19 +1127,20 @@ static ssize_t lc_socket_recvmsg_encoded(lc_socket_t *sock, struct msghdr *msg,
 		memcpy(msg->msg_iov[i].iov_base, ptr, msg->msg_iov[i].iov_len);
 		ptr += msg->msg_iov[i].iov_len;
 	}
-err_0:
+err_free_buf:
 	free(buf);
+	if (err) errno = err;
 	return bytes;
 }
 #endif
 
 ssize_t lc_socket_recvmsg(lc_socket_t *sock, struct msghdr *msg, int flags)
 {
-	if ((sock->ccoding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+	if (sock->ccoding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 		return lc_socket_recvmsg_encoded(sock, msg, flags);
 #else
-		errno = ENOTSUP; return -1;
+		return (errno = ENOTSUP), -1;
 #endif
 	}
 	if (sock->ifx) {
@@ -1009,14 +1161,15 @@ ssize_t lc_socket_recv(lc_socket_t *sock, void *buf, size_t len, int flags)
 		char ctl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
 		struct in6_pktinfo pi = {0};
 		struct msghdr msg = { .msg_iov = &iov, .msg_iovlen = 1 };
+
 		msg.msg_control = ctl;
 		msg.msg_controllen = sizeof ctl;
 		bytes = lc_socket_recvmsg_if(sock, &msg, flags, &pi);
-		if ((sock->ccoding & LC_CODE_SYMM) == LC_CODE_SYMM) {
+		if (sock->ccoding & LC_CODE_SYMM) {
 #ifdef HAVE_LIBSODIUM
 			return lc_socket_decoded(sock, &msg, bytes, &pi);
 #else
-			errno = ENOTSUP; return -1;
+			return (errno = ENOTSUP), -1;
 #endif
 		}
 		return bytes;
@@ -1027,29 +1180,51 @@ ssize_t lc_socket_recv(lc_socket_t *sock, void *buf, size_t len, int flags)
 #ifdef HAVE_LIBLCRQ
 inline static ssize_t lc_channel_recv_lcrq(lc_channel_t *chan, void *buf, size_t len, int flags)
 {
+	const uint64_t F = rq_F(chan->rq);
 	const uint16_t OH = RQ_OVERHEAD;
 	const uint16_t KP = rq_KP(chan->rq);
 	const uint16_t K = rq_K(chan->rq);
 	const uint16_t Z = rq_Z(chan->rq);
 	const uint16_t T = rq_T(chan->rq);
-	const uint16_t M = (KP + OH) * Z;
-	uint32_t ESI[Z][M];
-	uint8_t blk[Z][M * T];
+	const uint16_t nesi = KP + OH;
+	const uint32_t M = nesi * (uint32_t)Z;
+	uint32_t ESI[Z][nesi];
+	uint8_t *blk;
 	uint8_t sym[T];
+	uint8_t sbn;
+	uint32_t esi;
 	uint16_t orig = K;
 	uint16_t rec = 0;
-	struct iovec iov[2];
+	struct iovec iov[4];
 	struct msghdr msg = {
 		.msg_iov = iov,
-		.msg_iovlen = sizeof iov / sizeof iov[0]
 	};
+	time_t start_time;
+	int rec_diff;
 	ssize_t zi = 0, rc;
+	rq_oti_t oti = 0;
+	rq_scheme_t scheme = 0;
 	rq_pid_t pid = 0;
+	int v = 0;
+	int idx;
 
-	iov[0].iov_base = &pid;
-	iov[0].iov_len = sizeof pid;
-	iov[1].iov_base = sym;
-	iov[1].iov_len = sizeof sym;
+	if ((size_t)F > len) return (errno = ERANGE), -1;
+
+	blk = calloc(Z, (uint32_t)nesi * T);
+	if (!blk) return -1;
+	pthread_cleanup_push(free, blk);
+
+	if (chan->coding & LC_CODE_FEC_OTI) {
+		iov[v].iov_base = &oti;
+		iov[v++].iov_len = sizeof oti;
+		iov[v].iov_base = &scheme;
+		iov[v++].iov_len = sizeof scheme;
+	}
+	iov[v].iov_base = &pid;
+	iov[v++].iov_len = sizeof pid;
+	iov[v].iov_base = sym;
+	iov[v++].iov_len = sizeof sym;
+	msg.msg_iovlen = v;
 	memset(ESI, ~0, sizeof ESI);
 
 	/* store the received symbol:
@@ -1058,42 +1233,118 @@ inline static ssize_t lc_channel_recv_lcrq(lc_channel_t *chan, void *buf, size_t
 	 * - if ESI >= K and all slots >= K used, backfill K
 	 * - update bitmap, ESI list
 	 * - if received symbols >= KP + OH, decode and return */
+	start_time = time(NULL);
+	rec_diff = 0;
 	while (orig && rec < M) {
-		int idx;
+		/* data seems to stop arriving in the middle, with no new
+		 * MLD listener reports being sent, so that the sender does
+		 * not know we are still waiting; so have a timeout on each
+		 * packet calculated as:
+		 * 30 seconds on the initial packet
+		 * 10 seconds within a second of the start
+		 * 10 times the average packet interval after that
+		 * this means that the timeout will self-adjust to the
+		 * rate things are arriving; the caller will need to retry
+		 * this, so we return 0 with an errno of ETIMEDOUT and they
+		 * can check for that */
+		struct pollfd pfd;
+		int timeout, nready;
 recvnext:
+		pfd.fd = chan->sock->sock;
+		pfd.events = POLLIN;
+		if (rec <= rec_diff) {
+			timeout = 30000;
+		} else {
+			time_t now = time(NULL), diff = now - start_time;
+			if (diff < 2)
+				timeout = 10000;
+			else
+				timeout = diff * 10000 / (rec - rec_diff);
+			rec_diff = rec;
+			start_time = now;
+		}
+		if ((nready = poll(&pfd, 1, timeout)) < 1) {
+			if (nready < 0) {
+				if (errno == EINTR) continue;
+				goto err_free_blk;
+			}
+			/* timed out, try a part and rejoin */
+			if (lc_channel_part(chan) == -1) goto err_free_blk;
+			if (lc_channel_join(chan) == -1) goto err_free_blk;
+			continue;
+		}
 		rc = lc_socket_recvmsg(chan->sock, &msg, flags);
-		if (rc == -1) return -1;
+		if (rc == -1) { zi = -1; goto err_free_blk; }
+		if (chan->coding & LC_CODE_FEC_OTI) {
+			if (rq_oti_F(oti) != F || rq_oti_T(oti) != T) continue;
+		}
 		zi += rc;
-		uint8_t sbn = pid >> 24;
-		uint32_t esi = ntohl(pid & 0x00ffffff);
+		sbn = rq_pid2sbn(pid);
+		esi = rq_pid2esi(pid);
+		if (sbn >= Z) continue;
 		if (esi < K) {
-			if (ESI[sbn][esi] != UINT32_MAX) continue;
+			if (ESI[sbn][esi] == esi) continue;
 			idx = esi;
 			orig--;
 		}
-		else for (idx = M; idx; idx--) {
+		else for (idx = nesi - 1; idx; idx--) {
 			if (ESI[sbn][idx] == esi) goto recvnext;
 			if (ESI[sbn][idx] == UINT32_MAX) break;
 		}
+		if (ESI[sbn][idx] == UINT32_MAX) rec++;
 		ESI[sbn][idx] = esi;
-		memcpy(blk[sbn] + T * idx, sym, T);
-		rec++;
+		memcpy(&blk[(sbn * nesi + idx) * T], sym, T);
 	}
 
 	/* decode blocks */
-	uint8_t dec[T * Z];
+	size_t blocksz = (uint32_t)T * K, sz = len;
+	uint8_t *dec = buf, *enc = blk;
 	for (uint8_t sbn = 0; sbn < Z; sbn++) {
-		if (!orig) memcpy(buf, blk[sbn], len); /* original symbols, no decoding */
-		else rq_decode(chan->rq, dec, blk[sbn], ESI[sbn], KP + OH);
+		if (!orig) memcpy(dec, enc, MIN(sz, blocksz)); /* original symbols, no decoding */
+		else {
+			if (sz < blocksz) {
+				uint8_t * temp = malloc(blocksz);
+				if (! temp) {
+					zi = -1; break;
+				}
+				if (rq_decode(chan->rq, temp, enc, ESI[sbn], nesi) == -1) {
+					zi = -1; break;
+				}
+				memcpy(dec, temp, sz);
+				free(temp);
+			} else {
+				if (rq_decode(chan->rq, dec, enc, ESI[sbn], nesi) == -1) {
+					zi = -1; break;
+				}
+			}
+		}
+		dec += blocksz; enc += blocksz; sz -= blocksz;
 	}
-	if (orig) memcpy(buf, dec, len);
-
+err_free_blk:
+	pthread_cleanup_pop(1); /* free(blk) */
 	return zi;
 }
 
+/* peek at packet header to get OTI params */
+int lc_channel_oti_peek(lc_channel_t *chan, rq_oti_t *oti, rq_scheme_t *scheme)
+{
+	char buf[1280];
+	ssize_t rc;
+	size_t peekbyt = (chan->coding & LC_CODE_SYMM) ?
+		sizeof buf : sizeof *oti + sizeof *scheme;
+	rc = lc_socket_recv(chan->sock, buf, peekbyt, MSG_PEEK);
+	if (rc >= (ssize_t)(sizeof *oti + sizeof *scheme)) {
+		memcpy(oti, buf, sizeof(rq_oti_t));
+		memcpy(scheme, buf + sizeof *oti, sizeof *scheme);
+		if (rq_oti_T(*oti) != chan->T) return (errno = EBADMSG), -1;
+		chan->rq = rq_init(rq_oti_F(*oti), chan->T);
+	}
+	return rc;
+}
+
 static ssize_t lc_channel_recv_decoded(lc_channel_t *chan, void *buf, size_t len, int flags)
 {
-	chan->rq = rq_init(len, chan->T);
+	if (!chan->rq) chan->rq = rq_init(len, chan->T);
 	if (chan->rq) return lc_channel_recv_lcrq(chan, buf, len, flags);
 	return -1;
 }
@@ -1101,8 +1352,9 @@ static ssize_t lc_channel_recv_decoded(lc_channel_t *chan, void *buf, size_t len
 
 ssize_t lc_channel_recv(lc_channel_t *chan, void *buf, size_t len, int flags)
 {
+	if (!chan || !buf || !len) return (errno = EINVAL), -1;
 #ifdef HAVE_LIBLCRQ
-	if ((chan->coding & LC_CODE_FEC_RQ) == LC_CODE_FEC_RQ) {
+	if (chan->coding & LC_CODE_FEC_RQ) {
 		return lc_channel_recv_decoded(chan, buf, len, flags);
 	}
 #endif
@@ -1116,7 +1368,6 @@ static void process_msg(lc_socket_call_t *sc, lc_message_t *msg)
 
 	inet_ntop(AF_INET6, &msg->dst, msg->dstaddr, INET6_ADDRSTRLEN);
 	inet_ntop(AF_INET6, &msg->src, msg->srcaddr, INET6_ADDRSTRLEN);
-	msg->sockid = sc->sock->id;
 
 	/* update channel stats */
 	chan = lc_channel_by_address(sc->sock->ctx, &msg->dst);
@@ -1191,9 +1442,9 @@ static int lc_channel_membership_all(int sock, int opt, struct ipv6_mreq *req)
 
 	if (getifaddrs(&ifaddr) == -1) return -1;
 	for (ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
-		if ((ifa->ifa_flags & IFF_MULTICAST) != IFF_MULTICAST
-		  || ifa->ifa_addr == NULL
-		  || ifa->ifa_addr->sa_family != AF_INET6) continue;
+		if (!(ifa->ifa_flags & IFF_MULTICAST)
+		   || ifa->ifa_addr == NULL
+		   || ifa->ifa_addr->sa_family != AF_INET6) continue;
 
 		req->ipv6mr_interface = if_nametoindex(ifa->ifa_name);
 
@@ -1323,48 +1574,20 @@ int lc_channel_bind(lc_socket_t *sock, lc_channel_t *chan)
 	return rc;
 }
 
-static int lc_hashgroup(char *baseaddr, unsigned char *group, size_t len,
-		struct in6_addr *addr, unsigned int flags)
-{
-#ifndef HASH_TYPE
-	errno = ENOTSUP; return -1;
-#else
-	unsigned char hashgrp[HASHSIZE];
-	hash_state state;
-
-	hash_init(&state, NULL, 0, HASHSIZE);
-	hash_update(&state, (unsigned char *)group, len);
-	hash_update(&state, (unsigned char *)&flags, sizeof(flags));
-	hash_final(&state, hashgrp, HASHSIZE);
-
-	/* we have 112 bits (14 bytes) available for the group address
-	 * XOR the hashed group with the base multicast address */
-	if (inet_pton(AF_INET6, baseaddr, &addr->s6_addr) != 1)
-		return LC_ERROR_INVALID_BASEADDR;
-	for (int i = 2; i < 16; i++) {
-		addr->s6_addr[i] ^= hashgrp[i];
-	}
-
-	return 0;
-#endif
-}
-
 static lc_channel_t * lc_channel_ins(lc_ctx_t *ctx, lc_channel_t *chan)
 {
-	chan->next = ctx->chan_list;
-	ctx->chan_list = chan;
+	if (!pthread_mutex_lock(&chan->ctx->mtx)) {
+		chan->next = ctx->chan_list;
+		ctx->chan_list = chan;
+		pthread_mutex_unlock(&chan->ctx->mtx);
+	}
 	return chan;
 }
 
-static inline void lc_channel_setid(lc_channel_t *chan)
-{
-	chan->id = ++chan_id;
-}
-
 lc_channel_t * lc_channel_sidehash(lc_channel_t *base, unsigned char *key, size_t keylen)
 {
 #ifndef HASH_TYPE
-	errno = ENOTSUP; return NULL;
+	return (errno = ENOTSUP), NULL;
 #else
 	struct in6_addr *in;
 	unsigned char *ptr;
@@ -1396,7 +1619,6 @@ lc_channel_t * lc_channel_copy(lc_ctx_t *ctx, lc_channel_t *chan)
 	lc_channel_t *copy = calloc(1, sizeof(lc_channel_t));
 	if (!copy) return NULL;
 	copy->ctx = ctx;
-	lc_channel_setid(copy);
 	memcpy(&copy->sa, &chan->sa, sizeof(struct sockaddr_in6));
 	return lc_channel_ins(ctx, copy);
 }
@@ -1407,10 +1629,18 @@ lc_channel_t *lc_channel_init(lc_ctx_t *ctx, struct sockaddr_in6 *sa)
 	chan = calloc(1, sizeof(lc_channel_t));
 	if (!chan) return NULL;
 	chan->ctx = ctx;
+	chan->bps_out = ctx->bps_out;
+	chan->bps_in = ctx->bps_in;
+	chan->coding = ctx->coding;
+	chan->key = ctx->key;
+	chan->pek = ctx->pek;
+	chan->sek = ctx->sek;
+	chan->psk = ctx->psk;
+	chan->ssk = ctx->ssk;
+	clock_gettime(CLOCK_TAI, &chan->byt_reset);
 #ifdef HAVE_LIBLCRQ
 	chan->T = LC_DEFAULT_RQ_T;
 #endif
-	lc_channel_setid(chan);
 	memcpy(&chan->sa, sa, sizeof(struct sockaddr_in6));
 	return lc_channel_ins(ctx, chan);
 }
@@ -1425,16 +1655,45 @@ lc_channel_t *lc_channel_init_grp(lc_ctx_t *ctx, struct in6_addr *grp, short por
 	return lc_channel_init(ctx, &sa);
 }
 
+void lc_hashtoaddr(struct in6_addr *addr, unsigned char *hash, size_t hashlen, unsigned char flags)
+{
+	addr->s6_addr[0] = 0xff;  /* multicast */
+	addr->s6_addr[1] = flags; /* set flags */
+	memcpy(&addr->s6_addr[2], hash, MIN(hashlen, 14)); /* 112 bits (14 bytes) for hash */
+}
+
+inline static int lc_hashgroup(struct in6_addr *addr, unsigned char *group, size_t len,
+		unsigned char flags)
+
+{
+#ifndef HASH_TYPE
+	return (errno = ENOTSUP), -1;
+#else
+	unsigned char hash[14];
+	hash_generic(hash, 14, group, len);
+	lc_hashtoaddr(addr, hash, sizeof hash, flags);
+	return 0;
+#endif
+}
+
+lc_channel_t * lc_channel_hash(lc_ctx_t *ctx, unsigned char *hash, size_t hashlen, unsigned char flags,
+		short port)
+{
+	struct sockaddr_in6 sa = {
+		.sin6_family = AF_INET6,
+		.sin6_port = htons(port)
+	};
+	lc_hashtoaddr(&sa.sin6_addr, hash, hashlen, flags);
+	return lc_channel_init(ctx, &sa);
+}
+
 lc_channel_t * lc_channel_nnew(lc_ctx_t *ctx, unsigned char *s, size_t len)
 {
 	struct sockaddr_in6 sa = {
 		.sin6_family = AF_INET6,
 		.sin6_port = htons(LC_DEFAULT_PORT)
 	};
-
-	if (lc_hashgroup(DEFAULT_ADDR, s, len, &sa.sin6_addr, 0))
-		return NULL;
-
+	if (lc_hashgroup(&sa.sin6_addr, s, len, LC_DEFAULT_FLAGS)) return NULL;
 	return lc_channel_init(ctx, &sa);
 }
 
@@ -1466,46 +1725,59 @@ static int lc_socket_groups_free(lc_socket_t *sock)
 }
 #endif
 
-void lc_socket_close(lc_socket_t *sock)
+static void socket_close(lc_socket_t *sock)
 {
-	if (!sock) return;
-
-	lc_socket_listen_cancel(sock);
 #ifndef IPV6_MULTICAST_ALL
 	lc_socket_groups_free(sock);
 #endif
-
 	if (sock->sock) close(sock->sock);
 	lc_socket_t *prev = NULL;
 	for (lc_socket_t *p = sock->ctx->sock_list; p; p = p->next) {
-		if (p->id == sock->id) {
+		if (p == sock) {
 			if (prev) prev->next = p->next;
 			else sock->ctx->sock_list = p->next;
+			break;
 		}
 		prev = p;
 	}
+}
+
+void lc_socket_close(lc_socket_t *sock)
+{
+	int err = errno;
+	if (!sock) return;
+	if (sock->ctx && !pthread_mutex_lock(&sock->ctx->mtx)) {
+		lc_socket_listen_cancel(sock);
+		socket_close(sock);
+		pthread_mutex_unlock(&sock->ctx->mtx);
+	}
 	free(sock);
+	errno = err;
 }
 
 void lc_ctx_free(lc_ctx_t *ctx)
 {
+	int err = errno;
 	if (ctx) {
-		void *p, *h;
-		p = ctx->sock_list;
-		while (p) {
-			h = p;
-			p = ((lc_socket_t *)p)->next;
-			lc_socket_close(h);
+		void *p;
+		pthread_mutex_lock(&ctx->mtx);
+		while (ctx->sock_list) {
+			p = ctx->sock_list;
+			socket_close(p);
+			free(p);
 		}
-		p = ctx->chan_list;
-		while (p) {
-			h = p;
-			p = ((lc_channel_t *)p)->next;
-			lc_channel_free(h);
+		while (ctx->chan_list) {
+			p = ctx->chan_list;
+			channel_free(p);
+			free(p);
 		}
 		if (ctx->sock >= 0) close(ctx->sock);
+		if (ctx->mdex) mdex_free(ctx->mdex);
+		pthread_mutex_unlock(&ctx->mtx);
+		pthread_mutex_destroy(&ctx->mtx);
 		free(ctx);
 	}
+	errno = err;
 }
 
 int lc_socket_bind(lc_socket_t *sock, unsigned int ifx)
@@ -1525,42 +1797,42 @@ lc_socket_t * lc_socket_new(lc_ctx_t *ctx)
 	sock = calloc(1, sizeof(lc_socket_t));
 	if (!sock) return NULL;
 	sock->ctx = ctx;
-	sock->id = ++sock_id;
-	sock->next = ctx->sock_list;
-	ctx->sock_list = sock;
 	s = socket(AF_INET6, SOCK_DGRAM, 0);
 	if (s == -1) {
 		err = errno;
-		goto err_0;
+		goto err_free_sock;
 	}
 	sock->sock = s;
+	if (pthread_mutex_lock(&ctx->mtx)) goto err_free_sock;
+	sock->next = ctx->sock_list;
+	ctx->sock_list = sock;
+	pthread_mutex_unlock(&ctx->mtx);
 #ifdef IPV6_MULTICAST_ALL
 	/* available in Linux 4.2 onwards */
 	i = 0;
 	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_ALL, &i, sizeof i) == -1) {
-		goto err_1;
+		goto err_close_s;
 	}
 #endif
 	i = 1;
 	if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVPKTINFO, &i, sizeof i) == -1) {
-		goto err_1;
+		goto err_close_s;
 	}
 	i = DEFAULT_MULTICAST_LOOP;
 	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &i, sizeof i) == -1) {
-		goto err_1;
+		goto err_close_s;
 	}
 	i = DEFAULT_MULTICAST_HOPS;
 	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &i, sizeof i) == -1) {
-		goto err_1;
+		goto err_close_s;
 	}
-	return sock;
-err_1:
+	if (!(ctx->ifx && lc_socket_bind(sock, ctx->ifx))) return sock;
+err_close_s:
 	err = errno;
 	close(s);
-err_0:
+err_free_sock:
 	free(sock);
-	errno = err;
-	return NULL;
+	return (errno = err), NULL;
 }
 
 lc_ctx_t * lc_ctx_new(void)
@@ -1568,10 +1840,24 @@ lc_ctx_t * lc_ctx_new(void)
 	lc_ctx_t *ctx;
 
 	if (!(ctx = calloc(1, sizeof(lc_ctx_t)))) return NULL; /* errno set by calloc */
-	ctx->id = ++ctx_id;
 	ctx->next = ctx_list;
 	ctx_list = ctx;
 	ctx->sock = -1;
+	if (pthread_mutex_init(&ctx->mtx, NULL)) {
+		free(ctx); return NULL;
+	}
 
 	return ctx;
 }
+
+int lc_ctx_debug(lc_ctx_t *ctx, int debug)
+{
+	if (debug != -1) ctx->debug = debug;
+	return ctx->debug;
+}
+
+FILE *lc_ctx_stream(lc_ctx_t *ctx, FILE *stream)
+{
+	if (stream) ctx->stream = stream;
+	return ctx->stream;
+}
