@@ -1,22 +1,55 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2017-2022 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2017-2023 Brett Sheffield <bacs@librecast.net> */
 
 #ifndef _LIBRECAST_PVT_H
 #define _LIBRECAST_PVT_H 1
 
-#include "config.h"
-#include "../include/librecast/types.h"
+#include <config.h>
+#include <librecast/types.h>
 #include <pthread.h>
 #include <stddef.h>
 #ifdef HAVE_LIBLCRQ
-#include <lcrq.h>
+# include <lcrq.h>
 #endif
+#ifdef HAVE_MLD
+# include <mld.h>
+# include <librecast/q.h>
+#endif
+
+#ifdef HAVE_ENDIAN_H
+# include <endian.h>
+#elif defined(HAVE_SYS_ENDIAN_H)
+# include <sys/endian.h>
+#elif defined(HAVE_LIBKERN_OSBYTEORDER_H)
+# include <libkern/OSByteOrder.h>
+# define htobe64(x) OSSwapHostToBigInt64(x)
+# define be64toh(x) OSSwapBigToHostInt64(x)
+#endif
+
+/* fall back to CLOCK_REALTIME when Linux-specific CLOCK_TAI is unavailable */
+#ifndef CLOCK_TAI
+# define CLOCK_TAI CLOCK_REALTIME
+#endif
+
+#define NANO 1000000000
 
 typedef struct lc_ctx_t {
 	lc_ctx_t *next;
-	uint32_t id;
 	lc_socket_t *sock_list;
 	lc_channel_t *chan_list;
+	mdex_t *mdex;
+	pthread_mutex_t mtx; /* mutex for sock + chan lists */
+	lc_key_t key; /* symmetric key */
+	lc_key_t pek; /* public encryption key */
+	lc_key_t sek; /* secret encryption key */
+	lc_key_t psk; /* public signing key */
+	lc_key_t ssk; /* secret signing key */
+	FILE *stream;
+	int debug;
+	int coding;   /* encryption/encoding to use */
+	size_t bps_out; /* send rate limit (bps) */
+	size_t bps_in;  /* recv rate limit (bps) */
+	unsigned int ifx; /* default interface for sockets and channels */
 	int sock; /* AF_LOCAL socket for ioctls */
 } lc_ctx_t;
 
@@ -32,7 +65,6 @@ typedef struct lc_socket_t {
 	lc_socket_t *next;
 	lc_ctx_t *ctx;
 	pthread_t thread;
-	uint32_t id;
 	unsigned int ifx; /* interface index, 0 = all (default) */
 #ifndef IPV6_MULTICAST_ALL
 	lc_grplist_t *grps;
@@ -78,7 +110,11 @@ typedef struct lc_channel_t {
 	struct lc_socket_t *sock;
 	struct sockaddr_in6 sa;
 	char *uri;
-	uint32_t id;
+	size_t byt_out; /* bytes sent */
+	size_t byt_in;  /* bytes recv */
+	size_t bps_out; /* send rate limit (bps) */
+	size_t bps_in;  /* recv rate limit (bps) */
+	struct timespec byt_reset; /* byte counter reset */
 	lc_seq_t seq; /* sequence number (Lamport clock) */
 	lc_rnd_t rnd; /* random nonce */
 	lc_key_t key; /* symmetric key */
@@ -88,6 +124,7 @@ typedef struct lc_channel_t {
 	lc_key_t ssk; /* secret signing key */
 	int coding;   /* encryption/encoding to use */
 #ifdef HAVE_LIBLCRQ
+	rq_pid_t pid;
 	rq_t *rq;     /* RaptorQ context */
 	uint16_t T;   /* RaptorQ symbol size (bytes) */
 #endif
@@ -103,6 +140,32 @@ typedef struct lc_message_head_t {
 	lc_len_t len;
 } __attribute__((__packed__)) lc_message_head_t;
 
+struct lc_sync_options_s {
+	char *share;
+	size_t sharelen;
+	size_t trim;
+};
+
+#ifdef HAVE_MLD
+#include <semaphore.h>
+#include <librecast/mdex.h>
+struct qentry_s {
+	struct in6_addr grp;
+	mdex_entry_t entry;
+};
+struct lc_share_s {
+        lc_ctx_t *lctx;
+        mdex_t *mdex;
+	mld_t *mld;
+	mld_watch_t *watch;
+	pthread_t *tsend;
+	q_t q;
+	int nthread_send;
+        unsigned int ifx;
+	int flags;
+};
+#endif
+
 extern uint32_t ctx_id;
 extern uint32_t sock_id;
 extern uint32_t chan_id;
@@ -112,7 +175,7 @@ extern lc_socket_t *sock_list;
 extern lc_channel_t *chan_list;
 
 #define BUFSIZE 1500
-#define DEFAULT_ADDR "ff1e::"
+#define DEFAULT_FLAGS 0x1e
 #define SIDE_CHANNEL_NACK "NACK"
 
 #endif /* _LIBRECAST_PVT_H */
