@@ -1,6 +1,6 @@
 #include "test.h"
 
-int main()
+int main(void)
 {
 	result("new test");
 	return 0;
